# Copyright (c) Facebook, Inc. and its affiliates.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn


# class SimSiam(nn.Module):
#     """
#     Build a SimSiam model.
#     """
#     def __init__(self, base_encoder, dim=2048, pred_dim=512):
#         """
#         dim: feature dimension (default: 2048)
#         pred_dim: hidden dimension of the predictor (default: 512)
#         """
#         super(SimSiam, self).__init__()

#         # create the encoder
#         # num_classes is the output fc dimension, zero-initialize last BNs
#         #self.encoder = base_encoder(num_classes=dim, zero_init_residual=True)
#         self.encoder = base_encoder(num_classes=dim)

#         # build a 3-layer projector
#         prev_dim = self.encoder.fc.weight.shape[1]
#         self.encoder.fc = nn.Sequential(nn.Linear(prev_dim, prev_dim, bias=False),
#                                         nn.BatchNorm1d(prev_dim),
#                                         nn.ReLU(inplace=True), # first layer
#                                         nn.Linear(prev_dim, prev_dim, bias=False),
#                                         nn.BatchNorm1d(prev_dim),
#                                         nn.ReLU(inplace=True), # second layer
#                                         self.encoder.fc,
#                                         nn.BatchNorm1d(dim, affine=False)) # output layer
#         self.encoder.fc[6].bias.requires_grad = False # hack: not use bias as it is followed by BN

#         # build a 2-layer predictor
#         self.predictor = nn.Sequential(nn.Linear(dim, pred_dim, bias=False),
#                                         nn.BatchNorm1d(pred_dim),
#                                         nn.ReLU(inplace=True), # hidden layer
#                                         nn.Linear(pred_dim, dim)) # output layer

#     def forward(self, x1, x2):
#         """
#         Input:
#             x1: first views of images
#             x2: second views of images
#         Output:
#             p1, p2, z1, z2: predictors and targets of the network
#             See Sec. 3 of https://arxiv.org/abs/2011.10566 for detailed notations
#         """

#         # compute features for one view
#         z1 = self.encoder(x1) # NxC
#         z2 = self.encoder(x2) # NxC

#         p1 = self.predictor(z1) # NxC
#         p2 = self.predictor(z2) # NxC

#         return p1, p2, z1.detach(), z2.detach()

class SimSiam(nn.Module):
    """
    Build a SimSiam model.
    """
    def __init__(self, base_encoder, dim=2048, pred_dim=512):
        """
        dim: feature dimension (default: 2048)
        pred_dim: hidden dimension of the predictor (default: 512)
        """
        super(SimSiam, self).__init__()

        # create the encoder
        # num_classes is the output fc dimension, zero-initialize last BNs
        #self.encoder = base_encoder(num_classes=dim, zero_init_residual=True)
        self.encoder = base_encoder(num_classes=dim)
        #self.encoder = self.encoder.classifier = self.encoder.classifier[:-1]

        # build a 3-layer projector
        self.overlap = dim = 2048

        self.projector = nn.Sequential(nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap, affine=False))

        # build a 2-layer predictor
        self.predictor = nn.Sequential(nn.Linear(dim, pred_dim, bias=False),
                                        nn.BatchNorm1d(pred_dim),
                                        nn.ReLU(inplace=True), # hidden layer
                                        nn.Linear(pred_dim, dim)) # output layer

    def forward(self, x1, x2):
        """
        Input:
            x1: first views of images
            x2: second views of images
        Output:
            p1, p2, z1, z2: predictors and targets of the network
            See Sec. 3 of https://arxiv.org/abs/2011.10566 for detailed notations
        """

        # compute features for one view
        # compute features for one view
        f1 = self.encoder(x1) # NxC
        f2 = self.encoder(x2) # NxC

        z1 = self.projector(f1) # NxC
        z2 = self.projector(f2) # NxC

        p1 = self.predictor(z1) # NxC
        p2 = self.predictor(z2) # NxC

        return p1, p2, z1.detach(), z2.detach()



class SimSiamLSC(nn.Module):
    """
    Build a SimSiam model.
    """
    def __init__(self, base_encoder, dim=2048, pred_dim=512):
        """
        dim: feature dimension (default: 2048)
        pred_dim: hidden dimension of the predictor (default: 512)
        """
        super(SimSiamLSC, self).__init__()

        # create the encoder
        # num_classes is the output fc dimension, zero-initialize last BNs
        self.encoder = base_encoder(num_classes=dim, zero_init_residual=True)
        self.encoder.fc = nn.Identity()

        # build a 3-layer projector
        dimLSC = int(dim*(2/3))
        dim = dimLSC
        prev_dim = dimLSC#self.encoder.fc.weight.shape[1]  
        self.overlap = dimLSC     
        self.projector = nn.Sequential(nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap, affine=False))

        self.projectorPJ = nn.Sequential(nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap),
                                                        nn.ReLU(inplace=True),  # hidden layer
                                                        nn.Linear(self.overlap, self.overlap, bias=False),
                                                        nn.BatchNorm1d(self.overlap, affine=False))

        # build a 2-layer predictor       
        self.predictor = nn.Sequential(nn.Linear(dim, pred_dim, bias=False),
                                        nn.BatchNorm1d(pred_dim),
                                        nn.ReLU(inplace=True), # hidden layer
                                        nn.Linear(pred_dim, dim)) # output layer

        self.predictorPJ = nn.Sequential(nn.Linear(dim, pred_dim, bias=False),
                                        nn.BatchNorm1d(pred_dim),
                                        nn.ReLU(inplace=True), # hidden layer
                                        nn.Linear(pred_dim, dim)) # output layer

    def forward(self, x1, x2, x3, x4):
        """
        Input:
            x1: first views of images
            x2: second views of images
        Output:
            p1, p2, z1, z2: predictors and targets of the network
            See Sec. 3 of https://arxiv.org/abs/2011.10566 for detailed notations
        """

        # compute features for one view
        f1 = self.encoder(x1) # NxC
        f2 = self.encoder(x2) # NxC
        f3 = self.encoder(x3) # NxC
        f4 = self.encoder(x4) # NxC


        z1 = self.projector(f1[:, :self.overlap]) # NxC
        z2 = self.projector(f2[:, :self.overlap]) # NxC
        z3 = self.projectorPJ(f3[:, 2048-self.overlap:]) # NxC
        z4 = self.projectorPJ(f4[:, 2048-self.overlap:]) # NxC

        p1 = self.predictor(z1) # NxC
        p2 = self.predictor(z2) # NxC
        p3 = self.predictorPJ(z3) # NxC
        p4 = self.predictorPJ(z4) # NxC

        return p1, p2, p3, p4, z1.detach(), z2.detach(), z3.detach(), z4.detach()

