import csv
import os
import random

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

from data.dataloaders import get_test_data
from models.simsiam import FT_SimSiam
from utils.option_parser import optionParser

if __name__ == "__main__":

    ### Training Options

    torch.backends.cudnn.deterministic = True

    OP = optionParser()

    OP.parser = FT_SimSiam.add_model_specific_args(OP.parser)

    OP.parser.add_argument("--save_file", type=str, default="", help="filename to save")
    OP.parser.add_argument("--data_mode", type=str, default=None, help="test_data mode")

    args = OP.get_args()

    file_object = open("./results.txt", "a")

    ### cuda setup
    cuda_check = False
    if torch.cuda.is_available() and args.device != "cpu":
        cuda_check = True

    ### Seed setting for random operations
    seed = 1234
    torch.manual_seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    if cuda_check:
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    # data loaders
    print("\n===> Loading Dataset")

    testloader, n_classes = get_test_data(args)

    args.data_mode = "gray"
    testloader_gray, _ = get_test_data(args)

    # add model specific args.
    print("\n==> Model initialization")

    model = FT_SimSiam(args, n_classes)

    # checkpoint loading
    if args.ckpt != None:
        print("\n==> Resuming from checkpoint")
        checkpoint = torch.load(args.ckpt, map_location=lambda storage, loc: storage)
        if "model_state_dict" in checkpoint:
            model.load_pretrained(checkpoint["model_state_dict"])
        elif "state_dict" in checkpoint:
            model.load_pretrained(checkpoint["state_dict"])
        else:
            model.load_pretrained(checkpoint)

    ### GPU setting

    print("\n===> Setting GPU")

    if cuda_check:
        model.to(args.device)

    # receptive_field_dict = receptive_field(model.encoder, (3, 32, 32))

    ################################################################################
    ################################################################################
    ################################################################################

    print("\n===> Collecting data RGB\n")

    features = []
    with torch.no_grad():
        model.eval()
        for data in testloader:
            images, labels = data
            if cuda_check:
                images = images.to(args.device)
                labels = labels.to(args.device)

            features.append(model.get_features(images))

    features = torch.cat(features, 0)

    max_activations = features.max(2)[0].max(2)[0].cpu().numpy()

    names = testloader.dataset.data["File"].to_numpy()

    ################################################################################
    ### Grayscale

    print("\n===> Collecting data Gray\n")

    features_g = []
    with torch.no_grad():
        model.eval()
        for data in testloader_gray:
            images, labels = data
            if cuda_check:
                images = images.to(args.device)
                labels = labels.to(args.device)

            features_g.append(model.get_features(images))

    features_g = torch.cat(features_g, 0)

    max_activations_g = features_g.max(2)[0].max(2)[0].cpu().numpy()

    max_activations = np.sort(max_activations, 0)
    max_activations = max_activations[::-1, :]  # / max_activations.max()
    max_activations_g = np.sort(max_activations_g, 0)
    max_activations_g = max_activations_g[::-1, :]  # / max_activations_g.max()

    alphas = 1 - np.divide(
        np.sum(max_activations_g[0:100, :], axis=0),
        (np.sum(max_activations[0:100, :], axis=0) + 1e-8),
    )

    no_c = sum(alphas < 0.1)
    mid = sum(np.logical_and(alphas >= 0.1, alphas <= 0.25))
    col = sum(alphas > 0.25)

    print(
        "Neurons analysis:\n\tnon color selective: {}\n\tlow color selective: {}\n\thigh color selective: {}".format(
            no_c, mid, col
        )
    )

    np.savetxt(args.save_file, alphas, delimiter=",")
