import torch
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

from .transforms import PlanckianJitter, Instagram_filters


from data.cifar import CIFAR100
from data.datasets import (DataFromFolder, Flowers102,
                           sg568, TinyImageNet, Veg, CUB200,
                           T1Kplus, Paintings91, iNat1k)


DATA_DIR = "./datasets"

def get_training_data(args):

    if args.dataset.lower() == "cifar10":

        print("Dataset: Cifar 10")

        transform_train = transforms.Compose(
            [
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                ),
            ]
        )
        transform_test = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                ),
            ]
        )
        train_data = torchvision.datasets.CIFAR10(
            root="./data", train=True, download=True, transform=transform_train
        )
        trainloader = torch.utils.data.DataLoader(
            train_data,
            batch_size=args.batch_size,
            shuffle=True,
            num_workers=args.threads,
        )
        valid_data = torchvision.datasets.CIFAR10(
            root="./data", train=False, download=True, transform=transform_test
        )
        valloader = torch.utils.data.DataLoader(
            valid_data,
            batch_size=args.batch_size,
            shuffle=False,
            num_workers=args.threads,
        )

        n_classes = 10

    elif args.dataset.lower() == "cifar100":

        print("Dataset: Cifar 100")

        transform_train = transforms.Compose(
            [
                transforms.RandomCrop(32, padding=4),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                ),
            ]
        )
        transform_test = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                ),
            ]
        )
        train_data = CIFAR100(
            root="./data", train=True, download=True, transform=transform_train
        )
        trainloader = torch.utils.data.DataLoader(
            train_data,
            batch_size=args.batch_size,
            shuffle=True,
            num_workers=args.threads,
        )
        valid_data = CIFAR100(
            root="./data", train=False, download=True, transform=transform_test
        )
        valloader = torch.utils.data.DataLoader(
            valid_data,
            batch_size=args.batch_size,
            shuffle=False,
            num_workers=args.threads,
        )

        n_classes = 100

    elif args.dataset.lower() == "flowers102":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = f"{DATA_DIR}/102flowers"

        if args.from_scratch:
            print("Dataset: Flowers 102 from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: Flowers 102 fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            elif args.base_db == "imagenet":
                norm = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize(
                    [args.dim, args.dim]), norm]
            )

        train_data = Flowers102(im_dir, mode="train",
                                transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = Flowers102(
            im_dir, mode="test", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "tinyimagenet":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "../datasets/tiny-imagenet-200"

        if args.from_scratch:
            print("Dataset: Tiny-ImageNet from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([64, 64]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([64, 64])]
            )

        else:
            print("Dataset: Tiny-ImageNet fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            else:
                # raise Exception("Dataloader: base_db not defined.")
                def norm(x): return x

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([64, 64]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([64, 64]), norm]
            )

        train_data = TinyImageNet(
            im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = TinyImageNet(
            im_dir, mode="validation", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "vegfru":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = f"{DATA_DIR}/vegfru"

        if args.from_scratch:
            print("Dataset: VegFru from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: VegFru fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            elif args.base_db == "imagenet":
                norm = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize(
                    [args.dim, args.dim]), norm]
            )

        train_data = Veg(im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = Veg(im_dir, mode="validation", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "cub200":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = f"{DATA_DIR}/cubs_cropped"

        if args.from_scratch:
            print("Dataset: CUB200 from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: CUB200 fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            elif args.base_db == "imagenet":
                norm = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    # transforms.CenterCrop([args.dim, args.dim]),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                # [transforms.CenterCrop([args.dim, args.dim]),
                [transforms.ToTensor(),
                 transforms.Resize([args.dim, args.dim]),
                 norm]
            )

        train_data = CUB200(im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = CUB200(im_dir, mode="test", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "t1k":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = f"{DATA_DIR}/T1K+"

        if args.from_scratch:
            print("Dataset: T1K+ from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: T1K+ fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            elif args.base_db == "imagenet":
                norm = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    transforms.CenterCrop([args.dim, args.dim]),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    # transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                [transforms.CenterCrop([args.dim, args.dim]),
                 transforms.ToTensor(),
                 # transforms.Resize([args.dim, args.dim]),
                 norm]
            )

        train_data = T1Kplus(im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = T1Kplus(im_dir, mode="test", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "paintings91":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "../datasets/Paintings91"

        if args.from_scratch:
            print("Dataset: Paintings91 from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: Paintings91 fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    # transforms.CenterCrop([args.dim, args.dim]),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                # [transforms.CenterCrop([args.dim, args.dim]),
                [transforms.ToTensor(),
                 transforms.Resize([args.dim, args.dim]),
                 norm]
            )

        train_data = Paintings91(im_dir, mode="train",
                                 transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = Paintings91(im_dir, mode="test", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "inat1k":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "/fastsets/iNat"

        if args.from_scratch:
            print("Dataset: iNat1k from scratch")

            transform_train = transforms.Compose(
                [
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([args.dim, args.dim])]
            )

        else:
            print("Dataset: iNat1k fine tune from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                norm = transforms.Normalize(norm_mean, norm_std)
            elif args.base_db == "tinyimagenet":
                def norm(x): return x
            else:
                raise Exception("Dataloader: base_db not defined.")

            transform_train = transforms.Compose(
                [
                    # transforms.CenterCrop([args.dim, args.dim]),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm,
                ]
            )
            transform_val = transforms.Compose(
                [
                    # transforms.CenterCrop([args.dim, args.dim]),
                    transforms.ToTensor(),
                    transforms.Resize([args.dim, args.dim]),
                    norm
                ]
            )

        train_data = iNat1k(im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = iNat1k(im_dir, mode="validation",
                            transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    else:
        raise Exception("{} dataset not available.".format(args.dataset))

    print(
        "\n===> Dataset\n\nName: {}\nNumber of images train: {}\nNumber of images val: {}\nIterations per epoch: {}".format(
            train_data.__class__.__name__,
            len(train_data),
            len(valid_data),
            len(train_data) // args.batch_size,
        )
    )

    return trainloader, valloader, n_classes


################################################################################
################################################################################
################################################################################
# TEST DATALOADER

def get_test_data(args):

    if args.dataset.lower() == "cifar10":

        print("Dataset: Cifar 10")

        transform_test = transforms.Compose(
            [
                transforms.ToTensor(),
                transforms.Normalize(
                    (0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
                ),
            ]
        )
        valid_data = torchvision.datasets.CIFAR10(
            root="./data", train=False, download=True, transform=transform_test
        )
        valloader = torch.utils.data.DataLoader(
            valid_data,
            batch_size=args.batch_size,
            shuffle=False,
            num_workers=args.threads,
        )

        n_classes = 10

    elif args.dataset.lower() == "cifar100":

        print("Dataset: Cifar 100")

        if args.pj_analysis:
            transform_test = transforms.Compose(
                [
                    PlanckianJitter(mode="blackbody", idx=args.pj_idx),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                    ),
                ]
            )
        else:
            transform_test = transforms.Compose(
                [
                    Instagram_filters(),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        (0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
                    ),
                ]
            )

        valid_data = CIFAR100(
            root="./data", train=False, download=True, transform=transform_test
        )
        valloader = torch.utils.data.DataLoader(
            valid_data,
            batch_size=args.batch_size,
            shuffle=False,
            num_workers=args.threads,
        )

        n_classes = 100

    elif args.dataset.lower() == "flowers102":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "../datasets/102flowers"

        if args.from_scratch:
            print("Dataset: Flowers 102 from scratch")

            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([32, 32])]
            )

        else:
            print("Dataset: Flowers 102 fine tuned from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
            else:
                raise Exception("Dataloader: base_db not defined.")

            if args.pj_analysis:
                transform_val = transforms.Compose(
                    [
                        PlanckianJitter(mode="blackbody", idx=args.pj_idx),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )
            elif args.data_mode == "gray":
                transform_val = transforms.Compose(
                    [
                        transforms.Grayscale(num_output_channels=3),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        # transforms.Normalize(norm_mean, norm_std),
                    ]
                )
            else:
                transform_val = transforms.Compose(
                    [
                        Instagram_filters(),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )

        valid_data = Flowers102(
            im_dir, mode="validation", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    elif args.dataset.lower() == "tinyimagenet":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "../datasets/tiny-imagenet-200"

        if args.from_scratch:
            print("Dataset: Tiny-ImageNet from scratch")

            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([32, 32])]
            )

        else:
            print("Dataset: Tiny-ImageNet fine tuned from " + args.base_db)
            if args.base_db == "cifar10":
                norm_mean, norm_std = (
                    0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)
            elif args.base_db == "cifar100":
                norm_mean, norm_std = (
                    0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761)
            else:
                raise Exception("Dataloader: base_db not defined.")

            if args.pj_analysis:
                transform_val = transforms.Compose(
                    [
                        PlanckianJitter(mode="blackbody", idx=args.pj_idx),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )
            elif args.data_mode == "gray":
                transform_val = transforms.Compose(
                    [
                        transforms.Grayscale(num_output_channels=3),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        # transforms.Normalize(norm_mean, norm_std),
                    ]
                )
            else:
                transform_val = transforms.Compose(
                    [
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )

        valid_data = TinyImageNet(
            im_dir, mode="validation", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        n_classes = valid_data.n_classes

    else:
        raise Exception("{} dataset not available.".format(args.dataset))

    print(
        "\n===> Dataset\n\nName: {}\nNumber of images: {}".format(
            valid_data.__class__.__name__, len(valid_data)
        )
    )

    return valloader, n_classes


def get_set_from_dir(image_dir, opt):

    return DataLoader(
        dataset=DataFromFolder(image_dir),
        num_workers=opt.threads,
        batch_size=1,
        shuffle=False,
        pin_memory=False,
    )


def get_illuminant_est_datset(args):

    if args.dataset.lower() == "sg568":

        im_dir = args.im_dir
        if im_dir is None:
            im_dir = "../datasets/sg568"

        if args.from_scratch or args.test:
            print("Dataset:: Shi-Gehler Rep. from scratch")

            transform_train = transforms.Compose(
                [
                    # transforms.RandomResizedCrop([32, 32]),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Resize([32, 32]),
                ]
            )
            transform_val = transforms.Compose(
                [transforms.ToTensor(), transforms.Resize([32, 32])]
            )

        else:
            print("Dataset:: Shi-Gehler Rep. fine tune from " + args.base_db)
            if args.base_db != "":
                if args.base_db == "cifar10":
                    norm_mean, norm_std = (
                        (0.4914, 0.4822, 0.4465),
                        (0.2023, 0.1994, 0.2010),
                    )
                elif args.base_db == "cifar100":
                    norm_mean, norm_std = (
                        (0.5071, 0.4867, 0.4408),
                        (0.2675, 0.2565, 0.2761),
                    )
                else:
                    raise Exception("Dataloader:: base_db not defined.")

                transform_train = transforms.Compose(
                    [
                        # transforms.RandomResizedCrop([32, 32]),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )
                transform_val = transforms.Compose(
                    [
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                        transforms.Normalize(norm_mean, norm_std),
                    ]
                )
            else:
                print(
                    "Dataloader:: base db not defined: using no normalization in transforms!"
                )
                transform_train = transforms.Compose(
                    [
                        # transforms.RandomResizedCrop([32, 32]),
                        transforms.RandomHorizontalFlip(),
                        transforms.ToTensor(),
                        transforms.Resize([32, 32]),
                    ]
                )
                transform_val = transforms.Compose(
                    [transforms.ToTensor(), transforms.Resize([32, 32])]
                )

        train_data = sg568(im_dir, mode="train", transform=transform_train)
        trainloader = DataLoader(
            dataset=train_data,
            num_workers=args.threads,
            batch_size=args.batch_size,
            shuffle=True,
            pin_memory=True,
            drop_last=True,
        )

        valid_data = sg568(im_dir, mode="validation", transform=transform_val)
        valloader = DataLoader(
            dataset=valid_data,
            num_workers=args.threads,
            batch_size=1,  # args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

        test_data = sg568(im_dir, mode="test", transform=transform_val)
        testloader = DataLoader(
            dataset=test_data,
            num_workers=args.threads,
            batch_size=1,  # args.batch_size,
            shuffle=False,
            pin_memory=False,
            drop_last=False,
        )

    print(
        "\n===> Dataset\n\nName: {}\nNumber of images train: {}\nNumber of images val: {}\nNumber of images test: {}\nIterations per epoch: {}".format(
            train_data.__class__.__name__,
            len(train_data),
            len(valid_data),
            len(test_data),
            len(train_data) // args.batch_size,
        )
    )

    return trainloader, valloader, testloader
