from os import listdir
from os.path import join
import time

import numpy as np
import pandas
import skimage.io as io
import torch
import torch.utils.data as data
import torchvision.transforms as transforms
from PIL import Image


def is_image_file(filename):
    return any(
        filename.endswith(extension)
        for extension in [".png", ".jpg", ".jpeg", ".JPEG", ".bmp"]
    )


################################################################################
# 102 Flowers


class Flowers102(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(Flowers102, self).__init__()

        self.n_classes = 102
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/training.csv"
        elif mode == "validation":
            dts = image_dir + "/validation.csv"
        elif mode == "test":
            dts = image_dir + "/test.csv"
        else:
            raise Exception("Flowers102:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + "/jpg/" + self.data["File"][index])
        label = int(self.data["Label"][index]) - 1

        ### only luminance test ##################
        # img_ycbcr = np.array(inputt.convert("YCbCr"))
        #
        # img_ycbcr[:, :, 1] = img_ycbcr[:, :, 0]
        # img_ycbcr[:, :, 2] = img_ycbcr[:, :, 0]
        #
        # inputt = Image.fromarray(img_ycbcr)
        ##########################################

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class TinyImageNet(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(TinyImageNet, self).__init__()

        self.n_classes = 200
        self.im_dir = image_dir
        dts = None

        self.mode = mode
        if mode == "train":
            dts = image_dir + "/training.csv"
        elif mode == "validation":
            dts = image_dir + "/validation.csv"

        else:
            raise Exception("TinyImageNet:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + "/jpg/" + self.data["File"][index])
        if self.mode == "train":
            label = int(self.data["Label"][index])
        else:
            label = int(self.data["Label"][index]) - 1

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class VegFru(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(VegFru, self).__init__()

        self.n_classes = 292
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/vegfru_train.csv"
        elif mode == "validation":
            dts = image_dir + "/vegfru_val.csv"
        elif mode == "test":
            dts = image_dir + "/vegfru_test.csv"
        else:
            raise Exception("VegFru:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):
        inputt = Image.open(self.im_dir + "/" +
                            self.data["File"][index]).convert("RGB")
        label = int(self.data["Label"][index])

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class Veg(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(Veg, self).__init__()

        self.n_classes = 200
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/veg_train.csv"
        elif mode == "validation":
            dts = image_dir + "/veg_val.csv"
        elif mode == "test":
            dts = image_dir + "/veg_test.csv"
        else:
            raise Exception("Veg:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):
        inputt = Image.open(self.im_dir + "/" +
                            self.data["File"][index]).convert("RGB")
        label = int(self.data["Label"][index])

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class CUB200(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(CUB200, self).__init__()

        self.n_classes = 200
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/training.csv"
        elif mode == "test":
            dts = image_dir + "/test.csv"
        else:
            raise Exception("CUB200:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + "/images/" +
                            self.data["File"][index])
        label = int(self.data["Label"][index])

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class T1Kplus(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(T1Kplus, self).__init__()

        self.n_classes = 266
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/training.csv"
        elif mode == "test":
            dts = image_dir + "/test.csv"
        else:
            raise Exception("T1Kplus:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + "/images/" +
                            self.data["File"][index])
        label = int(self.data["Label"][index]) - 1

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class Paintings91(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(Paintings91, self).__init__()

        self.n_classes = 91
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/training_style.csv"
        elif mode == "test":
            dts = image_dir + "/test_style.csv"
        else:
            raise Exception("Paintings91:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + "/Images/" +
                            self.data["File"][index]).convert('RGB')
        label = int(self.data["Label"][index]) - 1

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]


class iNat1k(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(iNat1k, self).__init__()

        self.n_classes = 1010
        self.im_dir = image_dir
        dts = None
        if mode == "train":
            dts = image_dir + "/training.csv"
        elif mode == "validation":
            dts = image_dir + "/validation.csv"
        else:
            raise Exception("iNat1k:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(self.im_dir + '/train_val2019/' +
                            self.data["File"][index]).convert("RGB")
        label = int(self.data["Label"][index])

        input_g = self.trs(inputt)

        return input_g, label

    def __len__(self):
        return self.data.shape[0]

################################################################################
# From Folder


class DataFromFolder(data.Dataset):
    def __init__(self, image_dir):
        super(DataFromFolder, self).__init__()

        self.IN_filenames = [
            join(image_dir, x) for x in listdir(image_dir) if is_image_file(x)
        ]
        self.IN_filenames = sorted(self.IN_filenames)

        self.to_tensor = transforms.Compose(
            [transforms.ToTensor(), transforms.Resize([32, 32])]
        )

    def __getitem__(self, index):

        inputt = io.imread(self.IN_filenames[index])

        input_g = self.to_tensor(inputt)

        return input_g, self.IN_filenames[index]

    def __len__(self):
        return len(self.IN_filenames)


################################################################################
# Shi-Gehler


class sg568(data.Dataset):
    def __init__(self, image_dir, mode="train", transform=None):
        super(sg568, self).__init__()

        self.im_dir = image_dir
        self.mode = mode
        dts = None
        if mode == "train":
            dts = image_dir + "/train_sg568.csv"
        elif mode == "validation":
            dts = image_dir + "/val_sg568.csv"
        elif mode == "test":
            dts = image_dir + "/test_sg568.csv"
        else:
            raise Exception("sg568:: mode not defined")

        self.data = pandas.read_csv(dts)

        if transform is None:
            self.trs = transforms.Compose([transforms.ToTensor()])
        else:
            self.trs = transform

    def __getitem__(self, index):

        inputt = Image.open(
            self.im_dir + "/tiffraw/" + self.data["File"][index] + ".tiff"
        )
        inputt = np.array(inputt)
        mask = Image.open(self.im_dir + "/masks/" +
                          self.data["File"][index] + ".png")
        mask = np.array(mask)

        inputt = inputt * np.repeat(np.expand_dims(mask, 2), 3, 2)

        inputt = Image.fromarray(inputt)

        label = torch.FloatTensor(
            [self.data["R"][index], self.data["G"][index], self.data["B"][index]]
        )

        input_g = self.trs(inputt)

        if self.mode == "test":
            return input_g, label, self.data["File"][index]
        else:
            return input_g, label

    def __len__(self):
        return self.data.shape[0]
