import argparse
import json
from operator import is_
import os
from pprint import pprint
import random
import time
import warnings

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
from PIL import ImageFile
import numpy as np

from data.datasets import CUB200, Flowers102, T1Kplus, VegFru

from ray import tune

from data.transforms import PlanckianJitter
from main_lincls import LSC, DotDict, validate

# from data.dataloaders import get_training_data
ImageFile.LOAD_TRUNCATED_IMAGES = True



parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('-j', '--workers', default=12, type=int, metavar='N',
                    help='number of data loading workers (default: 8)')
parser.add_argument('--seed', default=1, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=0, type=int,
                    help='GPU id to use.')
parser.add_argument("--ckpt_dir", type=str)
parser.add_argument("--dim", type=int, default=224, help="image dim")


def parse_args():
    args = parser.parse_args()
    d = DotDict(args.__dict__)
    return d


DATA_DIR = "/data2/users/btwardow/pj/datasets"

def get_dateset(dataset, val_transforms):
    if dataset == "102flowers":
        im_dir = f"{DATA_DIR}/102flowers"
        val_dataset = Flowers102(im_dir, mode="test", transform=val_transforms)
    elif dataset == "cub200":
        im_dir = f"{DATA_DIR}/cubs_cropped"
        val_dataset = CUB200(im_dir, mode="test", transform=val_transforms)
    elif dataset == "vegfru":
        im_dir = f"{DATA_DIR}/vegfru"
        val_dataset = VegFru(im_dir, mode="validation", transform=val_transforms)
    elif dataset == "t1k":
        im_dir = f"{DATA_DIR}/T1K+"
        val_dataset = T1Kplus(im_dir, mode="test", transform=val_transforms)
    else:
        raise Exception(f"Unknow dataset: {dataset}")

    print('Test  dataset size:', len(val_dataset))
    return val_dataset

def lin_eval(args):
    print("ARGS:")
    pprint(args)

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        # torch.backends.cudnn.deterministic = True
        # warnings.warn('You have chosen to seed training. '
        #               'This will turn on the CUDNN deterministic setting, '
        #               'which can slow down your training considerably! '
        #               'You may see unexpected behavior when restarting '
        #               'from checkpoints.')

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu))

    # create models
    assert os.path.isdir(args.ckpt_dir)
    dataset = args.ckpt_dir.split('dataset=')[1].split(',')[0]
    method = args.ckpt_dir.split('method=')[1].split('_')[0]
    print('Dataset:', dataset)
    print('Method:', method)

    is_lsc = ('+' in method)
    print("=> creating model '{}'".format('resnet50'))
    model1 = models.__dict__['resnet50']()
    model1.fc = nn.Identity()

    model2 = None
    if is_lsc:
        model2 = models.__dict__['resnet50']()
        model2.fc = nn.Identity()



    # Data loading code
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])


    val_dataset = get_dateset(dataset, None) # just to get classes

    # init the fc layer
    encoder_dim = 2048
    if model2:
        encoder_dim *= 2
    lm = nn.Linear(encoder_dim, val_dataset.n_classes)
    lm.weight.data.normal_(mean=0.0, std=0.01)
    lm.bias.data.zero_()
    lm.cuda(args.gpu)

    if model2:
        model = LSC(model1, model2, lm)
    else:
        model1.fc = lm
        model = model1

    print('Loading the model')
    model.load_state_dict(torch.load(f"{args.ckpt_dir}/model.pt"), strict=True)
    model.cuda(args.gpu)

    results = []
    for pj_idx in range(25):
        print("Validating with PJ_idx:", pj_idx)

        val_transforms=transforms.Compose([
                transforms.Resize(256),
                transforms.CenterCrop(224),
                PlanckianJitter(idx=pj_idx),
                transforms.ToTensor(),
                normalize,
            ])

        val_dataset = get_dateset(dataset, val_transforms)
        val_loader = torch.utils.data.DataLoader(val_dataset,
            batch_size=256, shuffle=False,
            num_workers=args.workers, pin_memory=True)

        args.macro_acc = False
        args.print_freq = 10000
        acc1 = validate(val_loader, model, None, args)

        print('PJ idx: ', pj_idx, ' ACC:', acc1)
        results.append(acc1)

    np.savetxt(f"{args.ckpt_dir}/analysis.txt", np.array([results]), fmt='%f', delimiter=',')

if __name__ == '__main__':
    cli_args = parse_args()
    lin_eval(cli_args)
