import argparse
import json
from operator import is_
import os
from pprint import pprint
import random
import time
import warnings

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.distributed as dist
import torch.optim
import torch.multiprocessing as mp
import torch.utils.data
import torch.utils.data.distributed
import torchvision.transforms as transforms
import torchvision.datasets as datasets
import torchvision.models as models
from PIL import ImageFile
import numpy as np

from data.datasets import CUB200, Flowers102, T1Kplus, VegFru

from ray import tune

# from data.dataloaders import get_training_data
ImageFile.LOAD_TRUNCATED_IMAGES = True

model_names = sorted(name for name in models.__dict__
    if name.islower() and not name.startswith("__")
    and callable(models.__dict__[name]))


CKPT = {
    "PJ":"/data2/users/btwardow/pj/Experiments/SimSiam/ImageNet_224x224/plankian_checkpoint_0099.pth.tar",
    "CJ":"/data2/users/btwardow/pj/Experiments/SimSiam/ImageNet_224x224/simsiam_resnet50_checkpoint_0099.pth.tar",
    "CJ2":"/data2/users/btwardow/pj/Experiments/SimSiam/ImageNet_224x224/CJ_otherSeed_checkpoint_0099.pth.tar",
    "CJnoRG": "/data2/users/btwardow/pj/Experiments/SimSiam/ImageNet_224x224/CJ_noRG_checkpoint_0099.pth.tar",
}

parser = argparse.ArgumentParser(description='PyTorch ImageNet Training')
parser.add_argument('-a', '--arch', metavar='ARCH', default='resnet50',
                    choices=model_names,
                    help='model architecture: ' +
                        ' | '.join(model_names) +
                        ' (default: resnet50)')
parser.add_argument('-m', '--method', default='CJ', type=str, help='CSV embeddings to use, two -> LSC method')
parser.add_argument('-j', '--workers', default=12, type=int, metavar='N',
                    help='number of data loading workers (default: 8)')
parser.add_argument('--epochs', default=500, type=int, metavar='N',
                    help='number of total epochs to run')
parser.add_argument('-b', '--batch-size', default=256, type=int,
                    metavar='N',
                    help='mini-batch size (default: 4096), this is the total '
                         'batch size of all GPUs on the current node when '
                         'using Data Parallel or Distributed Data Parallel')
parser.add_argument('--lr', '--learning-rate', default=0.3, type=float,
                    metavar='LR', help='initial (base) learning rate', dest='lr')
parser.add_argument('--momentum', default=0.9, type=float, metavar='M',
                    help='momentum')
parser.add_argument('--wd', '--weight-decay', default=0., type=float,
                    metavar='W', help='weight decay (default: 0.)',
                    dest='weight_decay')
parser.add_argument('-p', '--print-freq', default=10, type=int,
                    metavar='N', help='print frequency (default: 10)')
parser.add_argument('--seed', default=1, type=int,
                    help='seed for initializing training. ')
parser.add_argument('--gpu', default=0, type=int,
                    help='GPU id to use.')
parser.add_argument("--dataset", type=str, default="102flowers", help="dataset to use")
parser.add_argument( "--optim", type=str, default="adam", help="optimizer to use")
parser.add_argument("--dim", type=int, default=224, help="image dim")
parser.add_argument('--lars', action='store_true', help='Use LARS')
parser.add_argument('--macro_acc', action='store_true')
parser.add_argument('--patience', type=int, default=30)

# Ray
parser.add_argument('--ray', action='store_true')
parser.add_argument('--ray_output_dir', type=str, default='ray_output/')
parser.add_argument('--ray_gpu_per_trial', type=float, default=0.5)
parser.add_argument('--ray_run_name', type=str, default='experiment_1')


class DotDict(dict):
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __init__(self, dct):
        for key, value in dct.items():
            if hasattr(value, 'keys'):
                value = DotDict(value)
            self[key] = value

class LSC(nn.Module):
    def __init__(self, model1, model2, lm):
        super().__init__()
        self.encoder1 = model1
        self.encoder2 = model2
        self.lm = lm
    def forward(self, inputs):
        features_1 = self.encoder1(inputs).detach()
        features_2 = self.encoder2(inputs).detach()
        features = torch.cat([features_1, features_2], 1)
        features = features.detach()
        outputs = self.lm(features)
        return outputs

def parse_args():
    args = parser.parse_args()
    d = DotDict(args.__dict__)
    return d


def lin_eval(args):

    if type(args) == dict:
        args = DotDict(args)

    print("ARGS:")
    pprint(args)

    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        # torch.backends.cudnn.deterministic = True
        # warnings.warn('You have chosen to seed training. '
        #               'This will turn on the CUDNN deterministic setting, '
        #               'which can slow down your training considerably! '
        #               'You may see unexpected behavior when restarting '
        #               'from checkpoints.')

    best_acc1 = 0
    best_acc1_epoch = 0
    best_model_state = None

    if args.gpu is not None:
        print("Use GPU: {} for training".format(args.gpu))

    # create models
    models_chkps = [CKPT[c] for c in args.method.split('+')]
    assert len(models_chkps) in [1, 2]
    print("=> creating model '{}'".format(args.arch))
    model1 = models.__dict__[args.arch]()
    model1.fc = nn.Identity()
    model2 = None
    if len(models_chkps) > 1:
        model2 = models.__dict__[args.arch]()
        model2.fc = nn.Identity()
        for _, param in model2.named_parameters():
            param.requires_grad = False

    # freeze all layers
    for _, param in model1.named_parameters():
        param.requires_grad = False


    # load from pre-trained
    for fn in models_chkps:
        assert os.path.isfile(fn)

    def _load_model_state(fn, model):
        print("=> loading checkpoint'{}'".format(fn))
        checkpoint = torch.load(fn, map_location="cpu")
        # rename moco pre-trained keys
        state_dict = checkpoint['state_dict']
        for k in list(state_dict.keys()):
            # retain only encoder up to before the embedding layer
            if k.startswith('module.encoder') and not k.startswith('module.encoder.fc'):
                # remove prefix
                state_dict[k[len("module.encoder."):]] = state_dict[k]
            # delete renamed or unused k
            del state_dict[k]

        msg = model.load_state_dict(state_dict, strict=False)
        print("missing keys:", msg.missing_keys )
        print("unexpected keys:", msg.unexpected_keys )
        print("=> loaded pre-trained model '{}'".format(fn))

    _load_model_state(models_chkps[0], model1)
    if model2:
        _load_model_state(models_chkps[1], model2)

    torch.cuda.set_device(args.gpu)
    model1.cuda(args.gpu)
    if model2:
        model2.cuda(args.gpu)

    # define loss function (criterion) and optimizer
    criterion = nn.CrossEntropyLoss().cuda(args.gpu)

    if args.lars:
        print("=> use LARS optimizer.")
        from apex.parallel.LARC import LARC
        optimizer = LARC(optimizer=optimizer, trust_coefficient=.001, clip=False)


    # Data loading code
    DATA_DIR = "/data2/users/btwardow/pj/datasets"
    normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406],
                                     std=[0.229, 0.224, 0.225])
    train_transforms = transforms.Compose([
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            normalize,
        ])
    val_transforms=transforms.Compose([
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            normalize,
        ])

    if args.dataset == "102flowers":
        im_dir = f"{DATA_DIR}/102flowers"
        train_dataset = Flowers102(im_dir, mode="train", transform=train_transforms)
        val_dataset = Flowers102(im_dir, mode="test", transform=val_transforms)
    elif args.dataset == "cub200":
        im_dir = f"{DATA_DIR}/cubs_cropped"
        train_dataset = CUB200(im_dir, mode="train", transform=train_transforms)
        val_dataset = CUB200(im_dir, mode="test", transform=val_transforms)
    elif args.dataset == "vegfru":
        im_dir = f"{DATA_DIR}/vegfru"
        train_dataset = VegFru(im_dir, mode="train", transform=train_transforms)
        val_dataset = VegFru(im_dir, mode="validation", transform=val_transforms)
    elif args.dataset == "t1k":
        im_dir = f"{DATA_DIR}/T1K+"
        train_dataset = T1Kplus(im_dir, mode="train", transform=train_transforms)
        val_dataset = T1Kplus(im_dir, mode="test", transform=val_transforms)
    else:
        raise Exception(f"Unknow dataset: {args.dataset}")

    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size,
        num_workers=args.workers, pin_memory=True)

    val_loader = torch.utils.data.DataLoader(val_dataset,
        batch_size=256, shuffle=False,
        num_workers=args.workers, pin_memory=True)

    print('Train dataset size:', len(train_dataset))
    print('Test  dataset size:', len(val_dataset))

    # init the fc layer
    encoder_dim = 2048
    if model2:
        encoder_dim *= 2
    lm = nn.Linear(encoder_dim, train_dataset.n_classes)
    lm.weight.data.normal_(mean=0.0, std=0.01)
    lm.bias.data.zero_()
    lm.cuda(args.gpu)

    if args.optim == "adam":
        optimizer = torch.optim.Adam(lm.parameters(), args.lr, weight_decay=args.weight_decay)
    else:
        optimizer = torch.optim.SGD(lm.parameters(), args.lr,
                                    momentum=args.momentum,
                                    weight_decay=args.weight_decay)

    if model2:
        model = LSC(model1, model2, lm)
    else:
        model1.fc = lm
        model = model1

    start_time = time.time()
    for epoch in range(args.epochs):
        adjust_learning_rate(optimizer, epoch, args)

        # train for one epoch
        train(train_loader, model, criterion, optimizer, epoch, args)

        # evaluate on validation set
        if args.macro_acc:
            acc1, macro_acc = validate(val_loader, model, criterion, args)
        else:
            acc1 = validate(val_loader, model, criterion, args)

        # remember best acc@1 and save checkpoint
        is_best = acc1 > best_acc1
        if is_best:
            best_acc1 = acc1
            best_acc1_epoch = epoch
            best_model_state = model.state_dict()
        # elif (epoch - best_acc1_epoch) > args.patience:
            # print('Patience is reached out. Stopping train...')
            # break

        if args.ray:
            end_time = time.time()
            time_taken_min = int((end_time - start_time)/60)
            if args.macro_acc:
                tune.report(best_acc1=best_acc1, best_acc1_epoch=best_acc1_epoch, macro_acc=macro_acc)
            else:
                tune.report(best_acc1=best_acc1, best_acc1_epoch=best_acc1_epoch)

    end_time = time.time()
    time_taken_min = int((end_time - start_time)/60)

    if args.ray:
        log_dir = tune.get_trial_dir()
    else:
        log_dir = 'experiments_2'

    with open(f'{log_dir}/config.json', 'wt') as f:
        json.dump(args, f)

    torch.save(best_model_state, f'{log_dir}/model.pt')

    with open(f'{log_dir}/results.json', 'wt') as f:
        json.dump({
            'time_taken_min': time_taken_min,
            'best_acc': best_acc1,
            'best_acc_epoch': best_acc1_epoch,
        }, f)


    print('Args:')
    pprint(args.__dict__)
    print('Time taken (min):', time_taken_min)
    print('Best ACC:', best_acc1)
    print('Best ACC epoch:', best_acc1_epoch)



def train(train_loader, model, criterion, optimizer, epoch, args):
    batch_time = AverageMeter('Time', ':6.3f')
    data_time = AverageMeter('Data', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(train_loader),
        [batch_time, data_time, losses, top1, top5],
        prefix="Epoch: [{}]".format(epoch))

    """
    Switch to eval mode:
    Under the protocol of linear classification on frozen features/models,
    it is not legitimate to change any part of the pre-trained model.
    BatchNorm in train mode may revise running mean/std (even if it receives
    no gradient), which are part of the model parameters too.
    """
    model.eval()

    end = time.time()
    for i, (images, target) in enumerate(train_loader):
        # measure data loading time
        data_time.update(time.time() - end)

        images = images.cuda(args.gpu, non_blocking=True)
        target = target.cuda(args.gpu, non_blocking=True)

        # compute output
        output = model(images)
        loss = criterion(output, target)

        # measure accuracy and record loss
        acc1, acc5 = accuracy(output, target, topk=(1, 5))
        losses.update(loss.item(), images.size(0))
        top1.update(acc1[0], images.size(0))
        top5.update(acc5[0], images.size(0))

        # compute gradient and do SGD step
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # measure elapsed time
        batch_time.update(time.time() - end)
        end = time.time()

        if i % args.print_freq == 0:
            progress.display(i)


def validate(val_loader, model, criterion, args):
    batch_time = AverageMeter('Time', ':6.3f')
    losses = AverageMeter('Loss', ':.4e')
    top1 = AverageMeter('Acc@1', ':6.2f')
    top5 = AverageMeter('Acc@5', ':6.2f')
    progress = ProgressMeter(
        len(val_loader),
        [batch_time, losses, top1, top5],
        prefix='Test: ')

    # switch to evaluate mode
    model.eval()

    per_class_counts = {}  # used in macro
    with torch.no_grad():
        end = time.time()
        for i, (images, target) in enumerate(val_loader):
            if args.gpu is not None:
                images = images.cuda(args.gpu, non_blocking=True)
            target = target.cuda(args.gpu, non_blocking=True)

            # compute output
            output = model(images)
            if criterion:
                loss = criterion(output, target)
            else:
                loss = 0.0

            if args.macro_acc:
                t_list = target.cpu().numpy().flatten().tolist()
                p_list = output.argmax(dim=1).cpu().numpy().flatten().tolist()
                for _t, _p in zip(t_list, p_list):
                    if _t not in per_class_counts:
                        per_class_counts[_t] = (0, 0)
                    _c, _a = per_class_counts[_t]
                    _a += 1
                    if _t == _p:
                        _c += 1
                    per_class_counts[_t] = (_c, _a)

            # measure accuracy and record loss
            acc1, acc5 = accuracy(output, target, topk=(1, 5))
            if criterion:
                losses.update(loss.item(), images.size(0))
            top1.update(acc1[0], images.size(0))
            top5.update(acc5[0], images.size(0))

            # measure elapsed time
            batch_time.update(time.time() - end)
            end = time.time()

            if i % args.print_freq == 0:
                progress.display(i)

    if args.macro_acc:
        # pprint(per_class_counts)
        macro_acc = np.mean([c/a for (c, a) in per_class_counts.values()]) * 100
        print(' * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f} Macro Acc@1 {macro_acc:.3f}'
                .format(top1=top1, top5=top5, macro_acc=macro_acc))
        return top1.avg.item(), macro_acc

    print(' * Acc@1 {top1.avg:.3f} Acc@5 {top5.avg:.3f}' .format(top1=top1, top5=top5))

    return top1.avg.item()


class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


class ProgressMeter(object):
    def __init__(self, num_batches, meters, prefix=""):
        self.batch_fmtstr = self._get_batch_fmtstr(num_batches)
        self.meters = meters
        self.prefix = prefix

    def display(self, batch):
        entries = [self.prefix + self.batch_fmtstr.format(batch)]
        entries += [str(meter) for meter in self.meters]
        print('\t'.join(entries))

    def _get_batch_fmtstr(self, num_batches):
        num_digits = len(str(num_batches // 1))
        fmt = '{:' + str(num_digits) + 'd}'
        return '[' + fmt + '/' + fmt.format(num_batches) + ']'


def adjust_learning_rate(optimizer, epoch, args):
    if args.optim != "adam":
        cur_lr = args.lr
        t1 = int(0.6 * args.epochs)
        t2 = int(0.8 * args.epochs)

        if epoch >= t1:
            cur_lr *= 0.1
        if epoch >= t2:
            cur_lr *= 0.1

        print('lr=', cur_lr)
        # """Decay the learning rate based on schedule"""
        # cur_lr = init_lr * 0.02 * (1. + math.cos(math.pi * epoch / args.epochs))
        for param_group in optimizer.param_groups:
            param_group['lr'] = cur_lr


def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    with torch.no_grad():
        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].reshape(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
        return res


CONFIG = {
    # "method" : tune.grid_search(['PJ', 'CJ', 'CJ+PJ']),
    "method" : tune.grid_search([ 'PJ', 'CJ', 'CJ+PJ', 'CJnoRG', 'CJ+CJnoRG']),
    # "dataset" : tune.grid_search(["cub200", "vegfru"]),
    # "dataset" : tune.grid_search(["vegfru"]),
    # "dataset" : tune.grid_search(["cub200"]),
    # "dataset" : tune.grid_search(["t1k"]),
    "dataset" : tune.grid_search(["vegfru", "t1k"]),
    # "dataset" : tune.grid_search(["102flowers"]),
    # "macro_acc": True,
}

# CONFIG = {
#     "method" : tune.grid_search(['CJnoRG', 'PJ+CJnoRG', 'CJ+CJnoRG', 'CJ2+CJnoRG']),
#     "dataset" : tune.grid_search(["cub200", "vegfru", "t1k"]),
# }


if __name__ == '__main__':
    cli_args = parse_args()

    print('CLI ARGS: ', cli_args)

    if cli_args.ray:
        print("Using Ray Tune...")
        config = cli_args.copy()
        config.update(CONFIG)

        print('RAY CONFIG:', config)

        os.environ['TUNE_DISABLE_AUTO_CALLBACK_LOGGERS'] = '1'

        metric_cols = ['best_acc1', 'best_acc1_epoch', 'macro_acc']
        if cli_args.macro_acc:
            metric_cols += ['macro_acc']
        reporter = tune.CLIReporter(max_progress_rows=100, metric_columns=metric_cols)

        analysis = tune.run(
            lin_eval,
            name=cli_args.ray_run_name,
            metric="best_acc1",
            mode="max",
            config=config,
            resources_per_trial = {
                "cpu" : 2,
                "gpu" : cli_args.ray_gpu_per_trial,
            },
            local_dir=cli_args.ray_output_dir,
            progress_reporter=reporter,
        )

        df = analysis.results_df
        df.to_csv(f"{cli_args.ray_output_dir}/{cli_args.ray_run_name}/results.csv")

    else:
        lin_eval(cli_args)
