import re
from argparse import ArgumentParser
from typing import Optional, Tuple

from typing import List, Tuple

import torch
from torch import nn
import torchvision
from models.squeezenet import squeezenet1_1
import pytorch_lightning as pl
from models.evaluator import SSLEvaluator
from models.resnets import resnet18, resnet50
from models.utils import torchvision_ssl_encoder
from utils.checkpoint_utils import load_pl_ckpt
from utils.weight_initializers import init_weights


class MLP(nn.Module):
    def __init__(
        self, input_dim: int = 2048, hidden_size: int = 4096, output_dim: int = 256
    ) -> None:
        super().__init__()
        self.output_dim = output_dim
        self.input_dim = input_dim
        self.model = nn.Sequential(
            nn.Linear(input_dim, hidden_size, bias=False),
            nn.BatchNorm1d(hidden_size),
            nn.ReLU(inplace=True),
            nn.Linear(hidden_size, output_dim, bias=True),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)
        return x


class SiameseArm(nn.Module):
    def __init__(
        self,
        encoder: Optional[nn.Module] = None,
        input_dim: int = 2048,
        hidden_size: int = 4096,
        output_dim: int = 256,
    ) -> None:
        super().__init__()

        if encoder is None:
            encoder = torchvision_ssl_encoder("resnet50")
        # Encoder
        self.encoder = encoder
        # Projector
        self.projector = MLP(input_dim, hidden_size, output_dim)
        # Predictor
        self.predictor = MLP(output_dim, hidden_size, output_dim)

    def forward(
        self, x: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        y = self.encoder(x)[0]
        z = self.projector(y)
        h = self.predictor(z)
        return y, z, h


class SimSiam(nn.Module):
    """
    PyTorch SimSiam implementation.
    Model from Pytorch Lightning implementation.
    """

    def __init__(
        self,
        arch: str = "resnet18",
        hidden_mlp: int = 512,
        feat_dim: int = 128,
        first_conv: bool = False,
        maxpool1: bool = False,
        **kwargs
    ):
        super().__init__()

        self.arch = arch

        self.hidden_mlp = hidden_mlp
        self.feat_dim = feat_dim
        self.first_conv = first_conv
        self.maxpool1 = maxpool1

        if self.arch == "resnet18":
            backbone = resnet18
        elif self.arch == "resnet50":
            backbone = resnet50

        encoder = backbone(
            first_conv=self.first_conv,
            maxpool1=self.maxpool1,
            return_all_feature_maps=False,
        )
        self.online_network = SiameseArm(
            encoder,
            input_dim=self.hidden_mlp,
            hidden_size=self.hidden_mlp,
            output_dim=self.feat_dim,
        )

    def forward(self, x):
        y, _, _ = self.online_network(x)
        return y

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        # model params
        parser.add_argument(
            "--arch", default="resnet50", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")
        parser.add_argument(
            "--hidden_mlp",
            default=2048,
            type=int,
            help="hidden layer dimension in projection head",
        )
        parser.add_argument(
            "--feat_dim", default=128, type=int, help="feature dimension"
        )
        parser.add_argument("--online_ft", action="store_true")
        parser.add_argument("--fp32", action="store_true")

        return parser


class Flatten(nn.Module):
    def __init__(self):
        super(Flatten, self).__init__()

    def forward(self, input_tensor):
        return input_tensor.view(input_tensor.size(0), -1)


class FT_SimSiam(nn.Module):
    """
    Fine_tune Model
    """

    def __init__(self, args, n_classes):
        super().__init__()

        self.model = args.model
        self.fs = args.from_scratch
        self.train_all = args.train_all

        # self.args = args
        if args.arch == "resnet18":
            backbone = resnet18
        elif args.arch == "resnet50":
            backbone = resnet50

        self.encoder = backbone(
            first_conv=args.first_conv,
            maxpool1=args.maxpool1,
            return_all_feature_maps=False,
        )

        self.evaluator = nn.Sequential(
            nn.Flatten(), nn.Linear(512, n_classes, bias=False)
        )

        init_weights(self.encoder, init_type="normal")
        init_weights(self.evaluator, init_type="normal")

    def forward(self, inputs):

        features = self.encoder(inputs)[0]

        if not self.fs and not self.train_all:
            features = features.detach()

        # - classifier for backbone
        outputs = self.evaluator(features)

        return outputs

    def get_features(self, inputs):

        with torch.no_grad():
            self.encoder.eval()
            features = self.encoder.get_features(inputs)[0]

        return features

    def load_pretrained(self, state_dict):

        if self.model == 'simsiam':
            new_dict = {
                k: state_dict[k] for k in state_dict if re.match(".*encoder.*", k)
            }
            new_dict = {re.sub(".*encoder.", "", k): v for k,
                        v in new_dict.items()}
            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)
        else:
            new_dict = {
                k: state_dict[k]
                for k in state_dict
                if re.match(".*resnet_barlowtwins.*", k)
            }

            pattern = re.compile(".*projection_mlp.*")
            matched_keys = {
                key for key, value in new_dict.items() if pattern.match(key)
            }
            for key in matched_keys:
                new_dict.pop(key, None)

            new_dict = {re.sub(".*backbone.", "", k): v for k, v in new_dict.items()}
            new_dict.pop("bn1.num_batches_tracked", None)

            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)

        if bool(new_dict):
            self.encoder.load_state_dict(new_dict)
        else:
            self.encoder.load_state_dict(state_dict)

        if not self.fs and not self.train_all:
            print("\nModel:: encoder parameters freezed.")
            for param in self.encoder.parameters():
                param.requires_grad = False

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        # model params
        parser.add_argument(
            "--arch", default="resnet18", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")

        return parser


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################


class SimSiam_mix(nn.Module):
    """
    Fine_tune Model
    """

    def __init__(self, args, n_classes):
        super().__init__()

        self.fs = args.from_scratch
        self.train_all = args.train_all
        self.model = args.model

        # self.args = args
        if args.arch == "resnet18":
            backbone = resnet18
        elif args.arch == "resnet50":
            backbone = resnet50

        self.encoder_1 = backbone(
            first_conv=args.first_conv,
            maxpool1=args.maxpool1,
            return_all_feature_maps=False,
        )

        self.encoder_2 = backbone(
            first_conv=args.first_conv,
            maxpool1=args.maxpool1,
            return_all_feature_maps=False,
        )

        self.evaluator = nn.Sequential(
            nn.Flatten(), nn.Linear(1024, n_classes, bias=False)
        )

        init_weights(self.encoder_1, init_type="normal")
        init_weights(self.encoder_2, init_type="normal")
        init_weights(self.evaluator, init_type="normal")

    def forward(self, inputs):

        features_1 = self.encoder_1(inputs)[0]
        features_2 = self.encoder_2(inputs)[0]

        # features = features_1
        features = torch.cat([features_1, features_2], 1)

        if not self.fs and not self.train_all:
            features = features.detach()

        # - classifier for backbone
        outputs = self.evaluator(features)

        return outputs

    def get_features(self, inputs):

        with torch.no_grad():
            self.encoder.eval()
            features = self.encoder.get_features(inputs)[0]

        return features

    def load_pretrained_1(self, state_dict):

        if self.model == 'simsiam':
            new_dict = {
                k: state_dict[k] for k in state_dict if re.match(".*encoder.*", k)
            }
            new_dict = {re.sub(".*encoder.", "", k): v for k,
                        v in new_dict.items()}
            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)
        else:
            new_dict = {
                k: state_dict[k]
                for k in state_dict
                if re.match(".*resnet_barlowtwins.*", k)
            }

            pattern = re.compile(".*projection_mlp.*")
            matched_keys = {
                key for key, value in new_dict.items() if pattern.match(key)
            }
            for key in matched_keys:
                new_dict.pop(key, None)

            new_dict = {re.sub(".*backbone.", "", k)                        : v for k, v in new_dict.items()}
            new_dict.pop("bn1.num_batches_tracked", None)

            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)

        if bool(new_dict):
            self.encoder_1.load_state_dict(new_dict)
        else:
            self.encoder_1.load_state_dict(state_dict)

        if not self.fs and not self.train_all:
            print("\nModel:: encoder parameters freezed.")
            for param in self.encoder_1.parameters():
                param.requires_grad = False

    def load_pretrained_2(self, state_dict):

        if self.model == 'simsiam':
            new_dict = {
                k: state_dict[k] for k in state_dict if re.match(".*encoder.*", k)
            }
            new_dict = {re.sub(".*encoder.", "", k): v for k,
                        v in new_dict.items()}
            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)
        else:
            new_dict = {
                k: state_dict[k]
                for k in state_dict
                if re.match(".*resnet_barlowtwins.*", k)
            }

            pattern = re.compile(".*projection_mlp.*")
            matched_keys = {
                key for key, value in new_dict.items() if pattern.match(key)
            }
            for key in matched_keys:
                new_dict.pop(key, None)

            new_dict = {re.sub(".*backbone.", "", k)                        : v for k, v in new_dict.items()}
            new_dict.pop("bn1.num_batches_tracked", None)

            new_dict.pop("fc.weight", None)
            new_dict.pop("fc.bias", None)

        if bool(new_dict):
            self.encoder_2.load_state_dict(new_dict)
        else:
            self.encoder_2.load_state_dict(state_dict)

        if not self.fs and not self.train_all:
            print("\nModel:: encoder parameters freezed.")
            for param in self.encoder_2.parameters():
                param.requires_grad = False

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        # model params
        parser.add_argument(
            "--arch", default="resnet18", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")

        return parser


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
################################################################################


class SimSiam_illest(nn.Module):
    """
    Model For illuminant estimation using simsiam architecture & weights
    """

    def __init__(self, args):
        super().__init__()

        self.model = args.model
        self.fs = args.from_scratch
        self.train_all = args.train_all

        # self.args = args
        if args.arch == "resnet18":
            backbone = resnet18
        if args.arch == "resnet18_nob":
            backbone = resnet18_nob
        elif args.arch == "resnet50":
            backbone = resnet50
        elif args.arch == "squeezenet":
            backbone = squeezenet1_1

        if (
            args.arch == "resnet18"
            or args.arch == "resnet18_nob"
            or args.arch == "resnet50"
        ):
            self.encoder = backbone(
                first_conv=args.first_conv,
                maxpool1=args.maxpool1,
                return_all_feature_maps=False,
            )
        else:
            self.encoder = backbone(pretrained=True)

        self.evaluator = nn.Sequential(
            nn.Flatten(), nn.Linear(512, 3, bias=False))

        if self.fs or self.train_all:
            init_weights(self.encoder, init_type="normal")
            init_weights(self.evaluator, init_type="normal")

    def load_pretrained(self, state_dict):

        import ipdb

        ipdb.set_trace()

        if self.model == 'simsiam':
            new_dict = {
                k: state_dict[k] for k in state_dict if re.match(".*encoder.*", k)
            }
            new_dict = {re.sub(".*encoder.", "", k): v for k,
                        v in new_dict.items()}
        else:
            import ipdb

            ipdb.set_trace()
            new_dict = {
                k: state_dict[k]
                for k in state_dict
                if re.match(".*resnet_barlowtwins.*", k)
            }
            new_dict = {re.sub(".*backbone.", "", k)                        : v for k, v in new_dict.items()}

        if bool(new_dict):
            self.encoder.load_state_dict(new_dict)
        else:
            self.encoder.load_state_dict(state_dict)

        if not self.fs and not self.train_all:
            print("\nModel:: encoder parameters freezed.")
            for param in self.encoder.parameters():
                param.requires_grad = False

    def freeze_encoder(self):
        print("\nModel:: encoder parameters freezed.")
        for param in self.encoder.parameters():
            param.requires_grad = False

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        # model params
        parser.add_argument(
            "--arch", default="resnet18", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")

        return parser

    def forward(self, inputs):

        features = self.encoder(inputs)

        if not self.fs and not self.train_all:
            features = features.detach()

        # - classifier for backbone
        outputs = self.evaluator(features)

        return torch.sigmoid(outputs)


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
################################################################################


class ProjectionHead(nn.Module):
    """Base class for all projection and prediction heads.
    Args:
        blocks:
            List of tuples, each denoting one block of the projection head MLP.
            Each tuple reads (in_features, out_features, batch_norm_layer,
            non_linearity_layer).
    Examples:
        >>> # the following projection head has two blocks
        >>> # the first block uses batch norm an a ReLU non-linearity
        >>> # the second block is a simple linear layer
        >>> projection_head = ProjectionHead([
        >>>     (256, 256, nn.BatchNorm1d(256), nn.ReLU()),
        >>>     (256, 128, None, None)
        >>> ])
    """

    def __init__(self, blocks: List[Tuple[int, int, nn.Module, nn.Module]]):

        super(ProjectionHead, self).__init__()

        self.layers = []
        for input_dim, output_dim, batch_norm, non_linearity in blocks:
            self.layers.append(nn.Linear(input_dim, output_dim))
            if batch_norm:
                self.layers.append(batch_norm)
            if non_linearity:
                self.layers.append(non_linearity)
        self.layers = nn.Sequential(*self.layers)

    def forward(self, x: torch.Tensor):
        """Computes one forward pass through the projection head.
        Args:
            x:
                Input of shape bsz x num_ftrs.
        """
        return self.layers(x)


class BarlowTwinsProjectionHead(ProjectionHead):
    """Projection head used for Barlow Twins.
    "The projector network has three linear layers, each with 8192 output
    units. The first two layers of the projector are followed by a batch
    normalization layer and rectified linear units." [0]
    [0]: 2021, Barlow Twins, https://arxiv.org/abs/2103.03230
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super(BarlowTwinsProjectionHead, self).__init__(
            [
                (input_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU()),
                (hidden_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU()),
                (hidden_dim, output_dim, None, None),
            ]
        )


class BarlowTwins(nn.Module):
    """Implementation of BarlowTwins[0] network.
    Recommended loss: :py:class:`lightly.loss.barlow_twins_loss.BarlowTwinsLoss`
    Default params are the ones explained in the original paper [0].
    [0] Zbontar,J. et.al. 2021. Barlow Twins... https://arxiv.org/abs/2103.03230
    Attributes:
        backbone:
            Backbone model to extract features from images.
            ResNet-50 in original paper [0].
        num_ftrs:
            Dimension of the embedding (before the projection head).
        proj_hidden_dim:
            Dimension of the hidden layer of the projection head. This should
            be the same size as `num_ftrs`.
        out_dim:
            Dimension of the output (after the projection head).
    """

    def __init__(
        self,
        backbone: nn.Module,
        num_ftrs: int = 2048,
        proj_hidden_dim: int = 8192,
        out_dim: int = 8192,
    ):

        super(BarlowTwins, self).__init__()

        self.backbone = backbone
        self.num_ftrs = num_ftrs
        self.proj_hidden_dim = proj_hidden_dim
        self.out_dim = out_dim

        self.projection_mlp = BarlowTwinsProjectionHead(
            num_ftrs, proj_hidden_dim, out_dim
        )

    def forward(
        self, x0: torch.Tensor, x1: torch.Tensor = None, return_features: bool = False
    ):
        """Forward pass through BarlowTwins.
        Extracts features with the backbone and applies the projection
        head to the output space. If both x0 and x1 are not None, both will be
        passed through the backbone and projection. If x1 is None, only x0 will
        be forwarded.
        Barlow Twins only implement a projection head unlike SimSiam.
        Args:
            x0:
                Tensor of shape bsz x channels x W x H.
            x1:
                Tensor of shape bsz x channels x W x H.
            return_features:
                Whether or not to return the intermediate features backbone(x).
        Returns:
            The output projection of x0 and (if x1 is not None)
            the output projection of x1. If return_features is
            True, the output for each x is a tuple (out, f) where f are the
            features before the projection head.
        Examples:
            >>> # single input, single output
            >>> out = model(x)
            >>>
            >>> # single input with return_features=True
            >>> out, f = model(x, return_features=True)
            >>>
            >>> # two inputs, two outputs
            >>> out0, out1 = model(x0, x1)
            >>>
            >>> # two inputs, two outputs with return_features=True
            >>> (out0, f0), (out1, f1) = model(x0, x1, return_features=True)
        """
        # forward pass first input
        f0 = self.backbone(x0).flatten(start_dim=1)
        out0 = self.projection_mlp(f0)

        # append features if requested
        if return_features:
            out0 = (out0, f0)

        if x1 is None:
            return out0

        # forward pass second input
        f1 = self.backbone(x1).flatten(start_dim=1)
        out1 = self.projection_mlp(f1)

        # append features if requested
        if return_features:
            out1 = (out1, f1)

        return out0, out1


class FT_BarlowTwins(nn.Module):
    """
    Fine_tune Model
    """

    def __init__(self, args, n_classes):
        super().__init__()

        self.fs = args.from_scratch
        self.train_all = args.train_all

        self.hidden_mlp = 512
        num_ftrs = 512  # from imagenette example

        # create a ResNet backbone and remove the classification head
        resnet = torchvision.models.resnet18()
        last_conv_channels = list(resnet.children())[-1].in_features
        self.backbone = nn.Sequential(
            *list(resnet.children()
                  )[:-1], nn.Conv2d(last_conv_channels, num_ftrs, 1)
        )
        # create a barlow twins model based on ResNet
        self.resnet_barlowtwins = BarlowTwins(
            self.backbone,
            num_ftrs=num_ftrs,
            proj_hidden_dim=2048,
            out_dim=self.hidden_mlp,
        )

        self.evaluator = nn.Sequential(
            nn.Flatten(), nn.Linear(self.hidden_mlp, n_classes, bias=False)
        )

        init_weights(self.evaluator, init_type="normal")

    def forward(self, inputs):

        features = self.resnet_barlowtwins(inputs)[0]

        if not self.fs and not self.train_all:
            features = features.detach()

        # - classifier for backbone
        outputs = self.evaluator(features)

        return outputs

    def get_features(self, inputs):

        with torch.no_grad():
            self.resnet_barlowtwins.eval()
            features = self.encoder.get_features(inputs)[0]

        return features

    def load_pretrained(self, state_dict):

        new_dict = {k: state_dict[k]
                    for k in state_dict if re.match(".*encoder.*", k)}
        new_dict = {re.sub(".*encoder.", "", k): v for k,
                    v in new_dict.items()}
        if bool(new_dict):
            self.resnet_barlowtwins.load_state_dict(new_dict)
        else:
            self.resnet_barlowtwins.load_state_dict(state_dict)

        if not self.fs and not self.train_all:
            print("\nModel:: encoder parameters freezed.")
            for param in self.encoder.parameters():
                param.requires_grad = False

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        # model params
        parser.add_argument(
            "--arch", default="resnet18", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")

        return parser
