import math
from argparse import ArgumentParser
from typing import Callable, Optional

from typing import List, Tuple

import numpy as np
import torch
import torchvision
import torchvision.transforms as transforms
from torch import nn, Tensor
from torch.nn import functional as F
from torch.optim.optimizer import Optimizer

import pytorch_lightning as pl
from pl_bolts.models.self_supervised.resnets import resnet50  # resnet18,

from models.resnets import resnet18
from models.squeezenet import squeezenet1_1
from pl_bolts.models.self_supervised.simsiam.models import SiameseArm
#from pl_bolts.optimizers.lars_scheduling import LARSWrapper
from pl_bolts.transforms.dataset_normalizations import (
    cifar10_normalization,
    imagenet_normalization,
    stl10_normalization,
)
from pytorch_lightning import seed_everything
from pytorch_lightning.loggers import TensorBoardLogger
from pytorch_lightning.utilities import AMPType


# warmup + decay as a function
def linear_warmup_decay(warmup_steps, total_steps, cosine=True, linear=False):
    """
    Linear warmup for warmup_steps, optionally with cosine annealing or
    linear decay to 0 at total_steps
    """
    assert not (linear and cosine)

    def fn(step):
        if step < warmup_steps:
            return float(step) / float(max(1, warmup_steps))

        if not (cosine or linear):
            # no decay
            return 1.0

        progress = float(step - warmup_steps) / float(
            max(1, total_steps - warmup_steps)
        )
        if cosine:
            # cosine decay
            return 0.5 * (1.0 + math.cos(math.pi * progress))

        # linear decay
        return 1.0 - progress

    return fn


################################################################################
################################################################################
################################################################################
################################################################################
# SimSiam
################################################################################


class SimSiam(pl.LightningModule):
    """
    PyTorch Lightning implementation of `Exploring Simple Siamese Representation Learning (SimSiam)
    <https://arxiv.org/pdf/2011.10566v1.pdf>`_
    Paper authors: Xinlei Chen, Kaiming He.
    Model implemented by:
        - `Zvi Lapp <https://github.com/zlapp>`_
    .. warning:: Work in progress. This implementation is still being verified.
    TODOs:
        - verify on CIFAR-10
        - verify on STL-10
        - pre-train on imagenet
    Example::
        model = SimSiam()
        dm = CIFAR10DataModule(num_workers=0)
        dm.train_transforms = SimCLRTrainDataTransform(32)
        dm.val_transforms = SimCLREvalDataTransform(32)
        trainer = pl.Trainer()
        trainer.fit(model, datamodule=dm)
    Train::
        trainer = Trainer()
        trainer.fit(model)
    CLI command::f
        # cifar10
        python simsiam_module.py --gpus 1
        # imagenet
        python simsiam_module.py
            --gpus 8
            --dataset imagenet2012
            --data_dir /path/to/imagenet/
            --meta_dir /path/to/folder/with/meta.bin/
            --batch_size 32
    """

    def __init__(
        self,
        gpus: int,
        num_samples: int,
        batch_size: int,
        dataset: str,
        num_nodes: int = 1,
        arch: str = "resnet50",
        hidden_mlp: int = 2048,
        feat_dim: int = 128,
        warmup_epochs: int = 10,
        max_epochs: int = 100,
        temperature: float = 0.1,
        first_conv: bool = True,
        maxpool1: bool = True,
        optimizer: str = "adam",
        lars_wrapper: bool = True,
        exclude_bn_bias: bool = False,
        start_lr: float = 0.0,
        learning_rate: float = 1e-3,
        final_lr: float = 0.0,
        weight_decay: float = 1e-6,
        **kwargs
    ):
        """
        Args:
            datamodule: The datamodule
            learning_rate: the learning rate
            weight_decay: optimizer weight decay
            input_height: image input height
            batch_size: the batch size
            num_workers: number of workers
            warmup_epochs: num of epochs for scheduler warm up
            max_epochs: max epochs for scheduler
        """
        super().__init__()
        self.save_hyperparameters()

        self.gpus = gpus
        self.num_nodes = num_nodes
        self.arch = arch
        self.dataset = dataset
        self.num_samples = num_samples
        self.batch_size = batch_size

        self.hidden_mlp = hidden_mlp
        self.feat_dim = feat_dim
        self.first_conv = first_conv
        self.maxpool1 = maxpool1

        self.optim = optimizer
        self.lars_wrapper = lars_wrapper
        self.exclude_bn_bias = exclude_bn_bias
        self.weight_decay = weight_decay
        self.temperature = temperature

        self.start_lr = start_lr
        self.final_lr = final_lr
        self.learning_rate = learning_rate
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs

        self.init_model()

        # compute iters per epoch
        nb_gpus = len(self.gpus) if isinstance(
            gpus, (list, tuple)) else self.gpus
        assert isinstance(nb_gpus, int)
        global_batch_size = (
            self.num_nodes * nb_gpus * self.batch_size
            if nb_gpus > 0
            else self.batch_size
        )
        self.train_iters_per_epoch = self.num_samples // global_batch_size

        # define LR schedule
        warmup_lr_schedule = np.linspace(
            self.start_lr,
            self.learning_rate,
            self.train_iters_per_epoch * self.warmup_epochs,
        )
        iters = np.arange(
            self.train_iters_per_epoch * (self.max_epochs - self.warmup_epochs)
        )
        cosine_lr_schedule = np.array(
            [
                self.final_lr
                + 0.5
                * (self.learning_rate - self.final_lr)
                * (
                    1
                    + math.cos(
                        math.pi
                        * t
                        / (
                            self.train_iters_per_epoch
                            * (self.max_epochs - self.warmup_epochs)
                        )
                    )
                )
                for t in iters
            ]
        )

        self.lr_schedule = np.concatenate(
            (warmup_lr_schedule, cosine_lr_schedule))

    def init_model(self):
        if self.arch == "resnet18":
            backbone = resnet18
        elif self.arch == "resnet50":
            backbone = resnet50
        elif self.arch == "squeezenet":
            backbone = squeezenet1_1

        if self.arch == "resnet18" or self.arch == "resnet50":
            encoder = backbone(
                first_conv=self.first_conv,
                maxpool1=self.maxpool1,
                return_all_feature_maps=False,
            )
        else:
            encoder = backbone(pretrained=False)

        self.online_network = SiameseArm(
            encoder,
            input_dim=self.hidden_mlp,
            hidden_size=self.hidden_mlp,
            output_dim=self.feat_dim,
        )

    def forward(self, x):
        y, _, _ = self.online_network(x)
        return y

    def cosine_similarity(self, a, b):
        b = b.detach()  # stop gradient of backbone + projection mlp
        a = F.normalize(a, dim=-1)
        b = F.normalize(b, dim=-1)
        sim = -1 * (a * b).sum(-1).mean()
        return sim

    def training_step(self, batch, batch_idx):

        (img_1, img_2, _), y = batch

        # Image 1 to image 2 loss
        _, z1, h1 = self.online_network(img_1)
        _, z2, h2 = self.online_network(img_2)

        loss = self.cosine_similarity(
            h1, z2) / 2 + self.cosine_similarity(h2, z1) / 2

        # log results
        self.log_dict({"loss": loss})

        return loss

    def validation_step(self, batch, batch_idx):

        (img_1, img_2, _), y = batch

        # Image 1 to image 2 loss
        _, z1, h1 = self.online_network(img_1)
        _, z2, h2 = self.online_network(img_2)
        loss = self.cosine_similarity(
            h1, z2) / 2 + self.cosine_similarity(h2, z1) / 2

        # log results
        self.log_dict({"loss": loss})

        return loss

    def exclude_from_wt_decay(
        self, named_params, weight_decay, skip_list=["bias", "bn"]
    ):
        params = []
        excluded_params = []

        for name, param in named_params:
            if not param.requires_grad:
                continue
            elif any(layer_name in name for layer_name in skip_list):
                excluded_params.append(param)
            else:
                params.append(param)

        return [
            {"params": params, "weight_decay": weight_decay},
            {"params": excluded_params, "weight_decay": 0.0},
        ]

    def configure_optimizers(self):
        if self.exclude_bn_bias:
            params = self.exclude_from_wt_decay(
                self.named_parameters(), weight_decay=self.weight_decay
            )
        else:
            params = self.parameters()

        if self.optim == "sgd":
            optimizer = torch.optim.SGD(
                params,
                lr=self.learning_rate,
                momentum=0.9,
                weight_decay=self.weight_decay,
            )
        elif self.optim == "adam":
            optimizer = torch.optim.Adam(
                params, lr=self.learning_rate, weight_decay=self.weight_decay
            )

        if self.lars_wrapper:
            optimizer = LARSWrapper(
                optimizer, eta=0.001, clip=False  # trust coefficient
            )

        return optimizer

    def optimizer_step(
        self,
        epoch: int,
        batch_idx: int,
        optimizer: Optimizer,
        optimizer_idx: int,
        optimizer_closure: Optional[Callable] = None,
        on_tpu: bool = False,
        using_native_amp: bool = False,
        using_lbfgs: bool = False,
    ) -> None:
        # warm-up + decay schedule placed here since LARSWrapper is not optimizer class
        # adjust LR of optim contained within LARSWrapper

        if self.lars_wrapper:
            for param_group in optimizer.optim.param_groups:
                param_group["lr"] = self.lr_schedule[self.trainer.global_step]
        else:
            for param_group in optimizer.param_groups:
                param_group["lr"] = self.lr_schedule[self.trainer.global_step]

        # log LR (LearningRateLogger callback doesn't work with LARSWrapper)
        self.log(
            "learning_rate",
            self.lr_schedule[self.trainer.global_step],
            on_step=True,
            on_epoch=False,
        )

        # from lightning
        if self.trainer.amp_backend == AMPType.NATIVE:
            optimizer_closure()
            self.trainer.scaler.step(optimizer)
        elif self.trainer.amp_backend == AMPType.APEX:
            optimizer_closure()
            optimizer.step()
        else:
            optimizer.step(closure=optimizer_closure)

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument(
            "--save_dir", default="tb_logs", type=str, help="convnet architecture"
        )
        # model params
        parser.add_argument(
            "--arch", default="resnet50", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")
        parser.add_argument(
            "--hidden_mlp",
            default=2048,
            type=int,
            help="hidden layer dimension in projection head",
        )
        parser.add_argument(
            "--feat_dim", default=128, type=int, help="feature dimension"
        )
        parser.add_argument("--online_ft", action="store_true")
        parser.add_argument("--fp32", action="store_true")

        # transform params
        parser.add_argument(
            "--gaussian_blur", action="store_true", help="add gaussian blur"
        )
        parser.add_argument(
            "--jitter_strength", type=float, default=1.0, help="jitter strength"
        )
        parser.add_argument(
            "--dataset", type=str, default="cifar10", help="stl10, cifar10"
        )
        parser.add_argument(
            "--data_dir", type=str, default=".", help="path to download data"
        )

        # training params
        parser.add_argument(
            "--num_workers", default=8, type=int, help="num of workers per GPU"
        )
        parser.add_argument(
            "--optimizer", default="adam", type=str, help="choose between adam/sgd"
        )
        parser.add_argument(
            "--lars_wrapper",
            action="store_true",
            help="apple lars wrapper over optimizer used",
        )
        parser.add_argument(
            "--exclude_bn_bias",
            action="store_true",
            help="exclude bn/bias from weight decay",
        )
        parser.add_argument(
            "--warmup_epochs", default=10, type=int, help="number of warmup epochs"
        )
        parser.add_argument(
            "--batch_size", default=128, type=int, help="batch size per gpu"
        )

        parser.add_argument(
            "--temperature",
            default=0.1,
            type=float,
            help="temperature parameter in training loss",
        )
        parser.add_argument(
            "--weight_decay", default=1e-6, type=float, help="weight decay"
        )
        parser.add_argument(
            "--learning_rate", default=1e-3, type=float, help="base learning rate"
        )
        parser.add_argument(
            "--start_lr", default=0, type=float, help="initial warmup learning rate"
        )
        parser.add_argument(
            "--final_lr", type=float, default=1e-6, help="final learning rate"
        )

        return parser


################################################################################
################################################################################
################################################################################
################################################################################
# SimCLR
################################################################################


class SyncFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, tensor):
        ctx.batch_size = tensor.shape[0]

        gathered_tensor = [
            torch.zeros_like(tensor) for _ in range(torch.distributed.get_world_size())
        ]

        torch.distributed.all_gather(gathered_tensor, tensor)
        gathered_tensor = torch.cat(gathered_tensor, 0)

        return gathered_tensor

    @staticmethod
    def backward(ctx, grad_output):
        grad_input = grad_output.clone()
        torch.distributed.all_reduce(
            grad_input, op=torch.distributed.ReduceOp.SUM, async_op=False
        )

        idx_from = torch.distributed.get_rank() * ctx.batch_size
        idx_to = (torch.distributed.get_rank() + 1) * ctx.batch_size
        return grad_input[idx_from:idx_to]


class Projection(nn.Module):
    def __init__(self, input_dim=2048, hidden_dim=2048, output_dim=128):
        super().__init__()
        self.output_dim = output_dim
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim

        self.model = nn.Sequential(
            nn.Linear(self.input_dim, self.hidden_dim),
            nn.BatchNorm1d(self.hidden_dim),
            nn.ReLU(),
            nn.Linear(self.hidden_dim, self.output_dim, bias=False),
        )

    def forward(self, x):
        x = self.model(x)
        return F.normalize(x, dim=1)


class SimCLR(pl.LightningModule):
    def __init__(
        self,
        gpus: int,
        num_samples: int,
        batch_size: int,
        dataset: str,
        num_nodes: int = 1,
        arch: str = "resnet50",
        hidden_mlp: int = 2048,
        feat_dim: int = 128,
        warmup_epochs: int = 10,
        max_epochs: int = 100,
        temperature: float = 0.1,
        first_conv: bool = True,
        maxpool1: bool = True,
        optimizer: str = "adam",
        exclude_bn_bias: bool = False,
        start_lr: float = 0.0,
        learning_rate: float = 1e-3,
        final_lr: float = 0.0,
        weight_decay: float = 1e-6,
        **kwargs
    ):
        """
        Args:
            batch_size: the batch size
            num_samples: num samples in the dataset
            warmup_epochs: epochs to warmup the lr for
            lr: the optimizer learning rate
            opt_weight_decay: the optimizer weight decay
            loss_temperature: the loss temperature
        """
        super().__init__()
        self.save_hyperparameters()

        self.gpus = gpus
        self.num_nodes = num_nodes
        self.arch = arch
        self.dataset = dataset
        self.num_samples = num_samples
        self.batch_size = batch_size

        self.hidden_mlp = hidden_mlp
        self.feat_dim = feat_dim
        self.first_conv = first_conv
        self.maxpool1 = maxpool1

        self.optim = optimizer
        self.exclude_bn_bias = exclude_bn_bias
        self.weight_decay = weight_decay
        self.temperature = temperature

        self.start_lr = start_lr
        self.final_lr = final_lr
        self.learning_rate = learning_rate
        self.warmup_epochs = warmup_epochs
        self.max_epochs = max_epochs

        self.encoder = self.init_model()

        self.projection = Projection(
            input_dim=self.hidden_mlp,
            hidden_dim=self.hidden_mlp,
            output_dim=self.feat_dim,
        )

        # compute iters per epoch
        global_batch_size = (
            self.num_nodes * self.gpus * self.batch_size
            if self.gpus > 0
            else self.batch_size
        )
        self.train_iters_per_epoch = self.num_samples // global_batch_size

    def init_model(self):
        if self.arch == "resnet18":
            backbone = resnet18
        elif self.arch == "resnet50":
            backbone = resnet50

        return backbone(
            first_conv=self.first_conv,
            maxpool1=self.maxpool1,
            return_all_feature_maps=False,
        )

    def forward(self, x):
        # bolts resnet returns a list
        return self.encoder(x)[-1]

    def shared_step(self, batch):
        if self.dataset == "stl10":
            unlabeled_batch = batch[0]
            batch = unlabeled_batch

        # final image in tuple is for online eval
        (img1, img2, _), y = batch

        # get h representations, bolts resnet returns a list
        h1 = self(img1)
        h2 = self(img2)

        # get z representations
        z1 = self.projection(h1)
        z2 = self.projection(h2)

        loss = self.nt_xent_loss(z1, z2, self.temperature)

        return loss

    def training_step(self, batch, batch_idx):
        loss = self.shared_step(batch)

        self.log("train_loss", loss, on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx):
        loss = self.shared_step(batch)

        self.log("val_loss", loss, on_step=False,
                 on_epoch=True, sync_dist=True)
        return loss

    def exclude_from_wt_decay(
        self, named_params, weight_decay, skip_list=["bias", "bn"]
    ):
        params = []
        excluded_params = []

        for name, param in named_params:
            if not param.requires_grad:
                continue
            elif any(layer_name in name for layer_name in skip_list):
                excluded_params.append(param)
            else:
                params.append(param)

        return [
            {"params": params, "weight_decay": weight_decay},
            {"params": excluded_params, "weight_decay": 0.0},
        ]

    def configure_optimizers(self):

        if self.exclude_bn_bias:
            params = self.exclude_from_wt_decay(
                self.named_parameters(), weight_decay=self.weight_decay
            )
        else:
            params = self.parameters()

        if self.optim == "sgd":
            optimizer = torch.optim.SGD(
                params,
                lr=self.learning_rate,
                momentum=0.9,
                weight_decay=self.weight_decay,
            )
        elif self.optim == "lars":
            optimizer = LARS(
                params,
                lr=self.learning_rate,
                momentum=0.9,
                weight_decay=self.weight_decay,
                trust_coefficient=0.001,
            )
        elif self.optim == "adam":
            optimizer = torch.optim.Adam(
                params, lr=self.learning_rate, weight_decay=self.weight_decay
            )

        warmup_steps = self.train_iters_per_epoch * self.warmup_epochs
        total_steps = self.train_iters_per_epoch * self.max_epochs

        scheduler = {
            "scheduler": torch.optim.lr_scheduler.CosineAnnealingLR(
                optimizer, total_steps
            ),
            # .LambdaLR(
            #     optimizer, linear_warmup_decay(warmup_steps, total_steps, cosine=True)
            # ),
            "interval": "step",
            "frequency": 1,
        }

        return [optimizer], [scheduler]

    def nt_xent_loss(self, out_1, out_2, temperature, eps=1e-6):
        """
        assume out_1 and out_2 are normalized
        out_1: [batch_size, dim]
        out_2: [batch_size, dim]
        """
        # gather representations in case of distributed training
        # out_1_dist: [batch_size * world_size, dim]
        # out_2_dist: [batch_size * world_size, dim]
        if torch.distributed.is_available() and torch.distributed.is_initialized():
            out_1_dist = SyncFunction.apply(out_1)
            out_2_dist = SyncFunction.apply(out_2)
        else:
            out_1_dist = out_1
            out_2_dist = out_2

        # out: [2 * batch_size, dim]
        # out_dist: [2 * batch_size * world_size, dim]
        out = torch.cat([out_1, out_2], dim=0)
        out_dist = torch.cat([out_1_dist, out_2_dist], dim=0)

        # cov and sim: [2 * batch_size, 2 * batch_size * world_size]
        # neg: [2 * batch_size]
        cov = torch.mm(out, out_dist.t().contiguous())
        sim = torch.exp(cov / temperature)
        neg = sim.sum(dim=-1)

        # from each row, subtract e^(1/temp) to remove similarity measure for x1.x1
        row_sub = Tensor(neg.shape).fill_(
            math.e ** (1 / temperature)).to(neg.device)
        # clamp for numerical stability
        neg = torch.clamp(neg - row_sub, min=eps)

        # Positive similarity, pos becomes [2 * batch_size]
        pos = torch.exp(torch.sum(out_1 * out_2, dim=-1) / temperature)
        pos = torch.cat([pos, pos], dim=0)

        loss = -torch.log(pos / (neg + eps)).mean()

        return loss

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)

        # model params
        parser.add_argument(
            "--arch", default="resnet50", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")
        parser.add_argument(
            "--hidden_mlp",
            default=2048,
            type=int,
            help="hidden layer dimension in projection head",
        )
        parser.add_argument(
            "--feat_dim", default=128, type=int, help="feature dimension"
        )
        parser.add_argument("--online_ft", action="store_true")
        parser.add_argument("--fp32", action="store_true")

        # transform params
        parser.add_argument(
            "--gaussian_blur", action="store_true", help="add gaussian blur"
        )
        parser.add_argument(
            "--jitter_strength", type=float, default=1.0, help="jitter strength"
        )
        parser.add_argument(
            "--dataset", type=str, default="cifar10", help="stl10, cifar10"
        )
        parser.add_argument(
            "--data_dir", type=str, default=".", help="path to download data"
        )

        # training params
        parser.add_argument("--fast_dev_run", default=1, type=int)
        parser.add_argument(
            "--num_nodes", default=1, type=int, help="number of nodes for training"
        )
        parser.add_argument(
            "--gpus", default=1, type=int, help="number of gpus to train on"
        )
        parser.add_argument(
            "--num_workers", default=8, type=int, help="num of workers per GPU"
        )
        parser.add_argument(
            "--optimizer", default="adam", type=str, help="choose between adam/lars"
        )
        parser.add_argument(
            "--exclude_bn_bias",
            action="store_true",
            help="exclude bn/bias from weight decay",
        )
        parser.add_argument(
            "--max_epochs", default=100, type=int, help="number of total epochs to run"
        )
        parser.add_argument("--max_steps", default=-1,
                            type=int, help="max steps")
        parser.add_argument(
            "--warmup_epochs", default=10, type=int, help="number of warmup epochs"
        )
        parser.add_argument(
            "--batch_size", default=128, type=int, help="batch size per gpu"
        )

        parser.add_argument(
            "--temperature",
            default=0.1,
            type=float,
            help="temperature parameter in training loss",
        )
        parser.add_argument(
            "--weight_decay", default=1e-6, type=float, help="weight decay"
        )
        parser.add_argument(
            "--learning_rate", default=1e-3, type=float, help="base learning rate"
        )
        parser.add_argument(
            "--start_lr", default=0, type=float, help="initial warmup learning rate"
        )
        parser.add_argument(
            "--final_lr", type=float, default=1e-6, help="final learning rate"
        )

        return parser


########################################################################################
########################################################################################
########################################################################################
########################################################################################
# BarlowTwins
########################################################################################


class BarlowTwinsLoss(torch.nn.Module):
    """Implementation of the Barlow Twins Loss from Barlow Twins[0] paper.
    This code specifically implements the Figure Algorithm 1 from [0].

    [0] Zbontar,J. et.al, 2021, Barlow Twins... https://arxiv.org/abs/2103.03230
        Examples:
        >>> # initialize loss function
        >>> loss_fn = BarlowTwinsLoss()
        >>>
        >>> # generate two random transforms of images
        >>> t0 = transforms(images)
        >>> t1 = transforms(images)
        >>>
        >>> # feed through SimSiam model
        >>> out0, out1 = model(t0, t1)
        >>>
        >>> # calculate loss
        >>> loss = loss_fn(out0, out1)
    """

    def __init__(self, lambda_param=5e-3):
        """Lambda param configuration with default value like in [0]
        Args:
            lambda_param ([float], optional): parameter for importance of
                redundancy reduction term.
            Defaults to 5e-3 [0].
        """
        super(BarlowTwinsLoss, self).__init__()
        self.lambda_param = lambda_param

    def forward(self, z_a: torch.Tensor, z_b: torch.Tensor):

        device = z_a.device

        # normalize repr. along the batch dimension
        z_a_norm = (z_a - z_a.mean(0)) / z_a.std(0)  # NxD
        z_b_norm = (z_b - z_b.mean(0)) / z_b.std(0)  # NxD

        N = z_a.size(0)
        D = z_a.size(1)

        # cross-correlation matrix
        c = torch.mm(z_a_norm.T, z_b_norm) / N  # DxD
        # loss
        c_diff = (c - torch.eye(D, device=device)).pow(2)  # DxD
        # multiply off-diagonal elems of c_diff by lambda
        c_diff[~torch.eye(D, dtype=bool)] *= self.lambda_param
        loss = c_diff.sum()

        return loss


class ProjectionHead(nn.Module):
    """Base class for all projection and prediction heads.
    Args:
        blocks:
            List of tuples, each denoting one block of the projection head MLP.
            Each tuple reads (in_features, out_features, batch_norm_layer,
            non_linearity_layer).
    Examples:
        >>> # the following projection head has two blocks
        >>> # the first block uses batch norm an a ReLU non-linearity
        >>> # the second block is a simple linear layer
        >>> projection_head = ProjectionHead([
        >>>     (256, 256, nn.BatchNorm1d(256), nn.ReLU()),
        >>>     (256, 128, None, None)
        >>> ])
    """

    def __init__(self, blocks: List[Tuple[int, int, nn.Module, nn.Module]]):

        super(ProjectionHead, self).__init__()

        self.layers = []
        for input_dim, output_dim, batch_norm, non_linearity in blocks:
            self.layers.append(nn.Linear(input_dim, output_dim))
            if batch_norm:
                self.layers.append(batch_norm)
            if non_linearity:
                self.layers.append(non_linearity)
        self.layers = nn.Sequential(*self.layers)

    def forward(self, x: torch.Tensor):
        """Computes one forward pass through the projection head.
        Args:
            x:
                Input of shape bsz x num_ftrs.
        """
        return self.layers(x)


class BarlowTwinsProjectionHead(ProjectionHead):
    """Projection head used for Barlow Twins.
    "The projector network has three linear layers, each with 8192 output
    units. The first two layers of the projector are followed by a batch
    normalization layer and rectified linear units." [0]
    [0]: 2021, Barlow Twins, https://arxiv.org/abs/2103.03230
    """

    def __init__(self, input_dim: int, hidden_dim: int, output_dim: int):
        super(BarlowTwinsProjectionHead, self).__init__(
            [
                (input_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU()),
                (hidden_dim, hidden_dim, nn.BatchNorm1d(hidden_dim), nn.ReLU()),
                (hidden_dim, output_dim, None, None),
            ]
        )


class BarlowTwins(nn.Module):
    """Implementation of BarlowTwins[0] network.
    Recommended loss: :py:class:`lightly.loss.barlow_twins_loss.BarlowTwinsLoss`
    Default params are the ones explained in the original paper [0].
    [0] Zbontar,J. et.al. 2021. Barlow Twins... https://arxiv.org/abs/2103.03230
    Attributes:
        backbone:
            Backbone model to extract features from images.
            ResNet-50 in original paper [0].
        num_ftrs:
            Dimension of the embedding (before the projection head).
        proj_hidden_dim:
            Dimension of the hidden layer of the projection head. This should
            be the same size as `num_ftrs`.
        out_dim:
            Dimension of the output (after the projection head).
    """

    def __init__(
        self,
        first_conv: bool = True,
        maxpool1: bool = True,
        arch: str = "resnet18",
        num_ftrs: int = 2048,
        proj_hidden_dim: int = 8192,
        out_dim: int = 8192,
    ):

        super(BarlowTwins, self).__init__()

        self.first_conv = first_conv
        self.maxpool1 = maxpool1
        self.arch = arch

        self.backbone = self.init_model()
        self.num_ftrs = num_ftrs
        self.proj_hidden_dim = proj_hidden_dim
        self.out_dim = out_dim

        self.projection_mlp = BarlowTwinsProjectionHead(
            num_ftrs, proj_hidden_dim, out_dim
        )

    def init_model(self):
        if self.arch == "resnet18":
            backbone = resnet18
        elif self.arch == "resnet50":
            backbone = resnet50

        return backbone(
            first_conv=self.first_conv,
            maxpool1=self.maxpool1,
            return_all_feature_maps=False,
        )

    def forward(
        self, x0: torch.Tensor, x1: torch.Tensor = None, return_features: bool = False
    ):
        """Forward pass through BarlowTwins.
        Extracts features with the backbone and applies the projection
        head to the output space. If both x0 and x1 are not None, both will be
        passed through the backbone and projection. If x1 is None, only x0 will
        be forwarded.
        Barlow Twins only implement a projection head unlike SimSiam.
        Args:
            x0:
                Tensor of shape bsz x channels x W x H.
            x1:
                Tensor of shape bsz x channels x W x H.
            return_features:
                Whether or not to return the intermediate features backbone(x).
        Returns:
            The output projection of x0 and (if x1 is not None)
            the output projection of x1. If return_features is
            True, the output for each x is a tuple (out, f) where f are the
            features before the projection head.
        Examples:
            >>> # single input, single output
            >>> out = model(x)
            >>>
            >>> # single input with return_features=True
            >>> out, f = model(x, return_features=True)
            >>>
            >>> # two inputs, two outputs
            >>> out0, out1 = model(x0, x1)
            >>>
            >>> # two inputs, two outputs with return_features=True
            >>> (out0, f0), (out1, f1) = model(x0, x1, return_features=True)
        """
        # forward pass first input
        f0 = self.backbone(x0)[-1].flatten(start_dim=1)
        out0 = self.projection_mlp(f0)

        # append features if requested
        if return_features:
            out0 = (out0, f0)

        if x1 is None:
            return out0

        # forward pass second input
        f1 = self.backbone(x1)[-1].flatten(start_dim=1)
        out1 = self.projection_mlp(f1)

        # append features if requested
        if return_features:
            out1 = (out1, f1)

        return out0, out1


class BarlowTwinsModel(pl.LightningModule):
    def __init__(self, args):
        super().__init__()

        self.max_epochs = args.max_epochs
        self.hidden_mlp = args.hidden_mlp
        self.arch = args.arch

        self.first_conv = args.first_conv
        self.maxpool1 = args.maxpool1

        num_ftrs = 512  # from imagenette example

        # create a ResNet backbone and remove the classification head
        # resnet = torchvision.models.resnet18()
        # last_conv_channels = list(resnet.children())[-1].in_features
        # self.backbone = nn.Sequential(
        #     *list(resnet.children())[:-1], nn.Conv2d(last_conv_channels, num_ftrs, 1)
        # )
        # create a barlow twins model based on ResNet
        self.resnet_barlowtwins = BarlowTwins(
            self.first_conv,
            self.maxpool1,
            self.arch,
            num_ftrs=num_ftrs,
            proj_hidden_dim=2048,
            out_dim=self.hidden_mlp,
        )
        self.criterion = BarlowTwinsLoss()

    def forward(self, x):
        return self.resnet_barlowtwins(x)

    def training_step(self, batch, batch_idx):
        (x0, x1, _), _ = batch
        # (img_1, img_2, _), y = batch
        x0, x1 = self.resnet_barlowtwins(x0, x1)
        loss = self.criterion(x0, x1)
        self.log_dict({"loss": loss})
        return loss

    def validation_step(self, batch, batch_idx):
        (x0, x1, _), _ = batch
        # (img_1, img_2, _), y = batch
        x0, x1 = self.resnet_barlowtwins(x0, x1)
        loss = self.criterion(x0, x1)
        self.log_dict({"loss": loss})
        return loss

    def configure_optimizers(self):
        optim = torch.optim.SGD(
            self.resnet_barlowtwins.parameters(),
            lr=6e-2,
            momentum=0.9,
            weight_decay=5e-4,
        )
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optim, self.max_epochs)
        return [optim], [scheduler]

    @staticmethod
    def add_model_specific_args(parent_parser):
        parser = ArgumentParser(parents=[parent_parser], add_help=False)
        parser.add_argument(
            "--save_dir", default="tb_logs", type=str, help="convnet architecture"
        )
        # model params
        parser.add_argument(
            "--arch", default="resnet50", type=str, help="convnet architecture"
        )
        # specify flags to store false
        parser.add_argument("--first_conv", action="store_false")
        parser.add_argument("--maxpool1", action="store_false")
        parser.add_argument(
            "--hidden_mlp",
            default=2048,
            type=int,
            help="hidden layer dimension in projection head",
        )
        parser.add_argument(
            "--feat_dim", default=128, type=int, help="feature dimension"
        )
        parser.add_argument("--online_ft", action="store_true")
        parser.add_argument("--fp32", action="store_true")

        # transform params
        parser.add_argument(
            "--gaussian_blur", action="store_true", help="add gaussian blur"
        )
        parser.add_argument(
            "--jitter_strength", type=float, default=1.0, help="jitter strength"
        )
        parser.add_argument(
            "--dataset", type=str, default="cifar10", help="stl10, cifar10"
        )
        parser.add_argument(
            "--data_dir", type=str, default=".", help="path to download data"
        )

        # training params
        parser.add_argument(
            "--num_workers", default=8, type=int, help="num of workers per GPU"
        )
        parser.add_argument(
            "--optimizer", default="adam", type=str, help="choose between adam/sgd"
        )
        parser.add_argument(
            "--lars_wrapper",
            action="store_true",
            help="apple lars wrapper over optimizer used",
        )
        parser.add_argument(
            "--exclude_bn_bias",
            action="store_true",
            help="exclude bn/bias from weight decay",
        )
        parser.add_argument(
            "--warmup_epochs", default=10, type=int, help="number of warmup epochs"
        )
        parser.add_argument(
            "--batch_size", default=128, type=int, help="batch size per gpu"
        )

        parser.add_argument(
            "--temperature",
            default=0.1,
            type=float,
            help="temperature parameter in training loss",
        )
        parser.add_argument(
            "--weight_decay", default=1e-6, type=float, help="weight decay"
        )
        parser.add_argument(
            "--learning_rate", default=1e-3, type=float, help="base learning rate"
        )
        parser.add_argument(
            "--start_lr", default=0, type=float, help="initial warmup learning rate"
        )
        parser.add_argument(
            "--final_lr", type=float, default=1e-6, help="final learning rate"
        )

        return parser


################################################################################
################################################################################
################################################################################


def cli_main():
    from pl_bolts.callbacks.ssl_online import SSLOnlineEvaluator
    from pl_bolts.datamodules import CIFAR10DataModule, ImagenetDataModule

    from data.cifar100_datamodule import CIFAR100DataModule
    from data.flower102_datamodule import Flower102DataModule

    # from pl_bolts.models.self_supervised.simclr import (
    #     SimCLREvalDataTransform,
    #         SimCLRTrainDataTransform,
    # )

    from data.simsiam_transforms import (
        SimSiamEvalDataTransform,
        SimSiamTrainDataTransform,
    )

    seed_everything(1234)

    parser = ArgumentParser()

    # trainer args
    parser = pl.Trainer.add_argparse_args(parser)

    # model args
    parser = BarlowTwinsModel.add_model_specific_args(parser)
    args = parser.parse_args()

    # pick data
    dm = None

    # init datamodule
    if args.dataset == "cifar10":
        val_split = 5000
        if args.num_nodes * args.gpus * args.batch_size > val_split:
            val_split = args.num_nodes * args.gpus * args.batch_size

        dm = CIFAR10DataModule(
            data_dir=args.data_dir,
            batch_size=args.batch_size,
            num_workers=args.num_workers,
            val_split=val_split,
            drop_last=True,
        )

        args.num_samples = dm.num_samples

        args.maxpool1 = False
        args.first_conv = False
        args.input_height = dm.size()[-1]
        args.temperature = 0.5

        normalization = cifar10_normalization()

        args.gaussian_blur = False
        args.jitter_strength = 0.5

    elif args.dataset == "cifar100":
        args.online_ft = True
        val_split = 5000
        if args.num_nodes * args.gpus * args.batch_size > val_split:
            val_split = args.num_nodes * args.gpus * args.batch_size

        dm = CIFAR100DataModule(
            data_dir=args.data_dir,
            batch_size=args.batch_size,
            num_workers=args.num_workers,
            val_split=val_split,
            drop_last=True,
        )

        args.num_samples = dm.num_samples

        args.maxpool1 = False
        args.first_conv = False
        args.input_height = dm.size()[-1]
        args.temperature = 0.5

        # ((0.5071, 0.4866, 0.4409), (0.2009, 0.1984, 0.2023))
        normalization = transforms.Normalize(
            mean=(0.5071, 0.4866, 0.4409), std=(0.2009, 0.1984, 0.2023)
        )

        args.gaussian_blur = False
        args.jitter_strength = 0.5

    elif args.dataset == "imagenet":
        args.maxpool1 = True
        args.first_conv = True
        normalization = imagenet_normalization()

        args.gaussian_blur = True
        args.jitter_strength = 1.0

        args.batch_size = 64
        args.num_nodes = 8
        args.gpus = 8  # per-node
        args.max_epochs = 800

        args.optimizer = "sgd"
        args.lars_wrapper = True
        args.learning_rate = 4.8
        args.final_lr = 0.0048
        args.start_lr = 0.3
        args.online_ft = True

        dm = ImagenetDataModule(
            data_dir=args.data_dir,
            batch_size=args.batch_size,
            num_workers=args.num_workers,
        )

        args.num_samples = dm.num_samples
        args.input_height = dm.size()[-1]

    elif args.dataset == "flower102":

        val_split = 5000
        if args.num_nodes * args.gpus * args.batch_size > val_split:
            val_split = args.num_nodes * args.gpus * args.batch_size

        dm = Flower102DataModule(
            batch_size=args.batch_size, num_workers=args.num_workers, drop_last=True
        )

        args.num_samples = dm.num_samples

        args.maxpool1 = False
        args.first_conv = False

        args.input_height = 32
        args.temperature = 0.5

        # ((0.5071, 0.4866, 0.4409), (0.2009, 0.1984, 0.2023))
        normalization = None
        # transforms.Normalize(
        #     mean=(0.5071, 0.4866, 0.4409), std=(0.2009, 0.1984, 0.2023)
        # )

        args.gaussian_blur = False
        # args.jitter_strength = 0.5

    else:
        raise NotImplementedError(
            "other datasets have not been implemented till now")

    dm.train_transforms = SimSiamTrainDataTransform(
        input_height=args.input_height,
        gaussian_blur=args.gaussian_blur,
        jitter_strength=args.jitter_strength,
        normalize=normalization,
    )

    dm.val_transforms = SimSiamEvalDataTransform(
        input_height=args.input_height,
        gaussian_blur=args.gaussian_blur,
        jitter_strength=args.jitter_strength,
        normalize=normalization,
    )

    model = SimSiam(**args.__dict__)
    # model = BarlowTwinsModel(args)

    # ckpt = torch.load(
    #     "./tb_logs/SimSiam_cifar100/version_22/checkpoints/epoch=799-step=69599.ckpt",
    #     map_location="cpu",
    # )
    # ckpt["state_dict"].pop("non_linear_evaluator.block_forward.2.weight")
    # ckpt["state_dict"].pop("non_linear_evaluator.block_forward.2.bias")
    # torch.save(ckpt, "./tmp.pt")
    #
    # model = SimSiam.load_from_checkpoint("./tmp.pt")

    # finetune in real-time
    online_evaluator = None
    if args.online_ft:
        # online eval
        # evaluation on downstream task
        online_evaluator = SSLOnlineEvaluator(
            drop_p=0.0,
            hidden_dim=None,
            z_dim=args.hidden_mlp,
            num_classes=dm.num_classes,
            dataset=args.dataset,
        )

    args.gpus = "0"
    trainer = pl.Trainer.from_argparse_args(
        args,
        sync_batchnorm=False,
        callbacks=[online_evaluator] if args.online_ft else None,
        default_root_dir="./checkpoints/",
        # check_val_every_n_epoch=500,
    )

    logger = TensorBoardLogger(
        args.save_dir, name="BarlowTwins_" + args.dataset)
    trainer.logger = logger

    trainer.fit(model, datamodule=dm)

class LARSWrapper(object):
    """
    Wrapper that adds LARS scheduling to any optimizer. This helps stability with huge batch sizes.
    References:
    - https://github.com/NVIDIA/apex/blob/master/apex/parallel/LARC.py
    - https://arxiv.org/pdf/1708.03888.pdf
    - https://github.com/noahgolmant/pytorch-lars/blob/master/lars.py
    """

    def __init__(self, optimizer, eta=0.02, clip=True, eps=1e-8):
        """
        Args:
            optimizer: torch optimizer
            eta: LARS coefficient (trust)
            clip: True to clip LR
            eps: adaptive_lr stability coefficient
        """
        self.optim = optimizer
        self.eta = eta
        self.eps = eps
        self.clip = clip

        # transfer optim methods
        self.state_dict = self.optim.state_dict
        self.load_state_dict = self.optim.load_state_dict
        self.zero_grad = self.optim.zero_grad
        self.add_param_group = self.optim.add_param_group
        self.__setstate__ = self.optim.__setstate__
        self.__getstate__ = self.optim.__getstate__
        self.__repr__ = self.optim.__repr__

    @property
    def defaults(self):
        return self.optim.defaults

    @defaults.setter
    def defaults(self, defaults):
        self.optim.defaults = defaults

    @property
    def __class__(self):
        return Optimizer

    @property
    def state(self):
        return self.optim.state

    @property
    def param_groups(self):
        return self.optim.param_groups

    @param_groups.setter
    def param_groups(self, value):
        self.optim.param_groups = value

    @torch.no_grad()
    def step(self, closure=None):
        weight_decays = []

        for group in self.optim.param_groups:
            weight_decay = group.get('weight_decay', 0)
            weight_decays.append(weight_decay)

            # reset weight decay
            group['weight_decay'] = 0

            # update the parameters
            [self.update_p(p, group, weight_decay) for p in group['params'] if p.grad is not None]

        # update the optimizer
        self.optim.step(closure=closure)

        # return weight decay control to optimizer
        for group_idx, group in enumerate(self.optim.param_groups):
            group['weight_decay'] = weight_decays[group_idx]

    def update_p(self, p, group, weight_decay):
        # calculate new norms
        p_norm = torch.norm(p.data)
        g_norm = torch.norm(p.grad.data)

        if p_norm != 0 and g_norm != 0:
            # calculate new lr
            new_lr = (self.eta * p_norm) / (g_norm + p_norm * weight_decay + self.eps)

            # clip lr
            if self.clip:
                new_lr = min(new_lr / group['lr'], 1)

            # update params with clipped lr
            p.grad.data += weight_decay * p.data
            p.grad.data *= new_lr


if __name__ == "__main__":
    cli_main()
