import csv
import math
import os
import random
import statistics as st

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

from data.dataloaders import get_illuminant_est_datset
from models.simsiam import SimSiam_illest
from utils.ColorMetrics import RecoveryLoss
from utils.option_parser import optionParser

if __name__ == "__main__":

    ### Training Options

    torch.backends.cudnn.deterministic = True

    OP = optionParser()

    OP.parser = SimSiam_illest.add_model_specific_args(OP.parser)

    args = OP.get_args()

    ### cuda setup
    cuda_check = False
    if torch.cuda.is_available() and args.device != "cpu":
        cuda_check = True

    ## Save dir setup

    save_path = args.save_dir

    save_paths = [save_path]

    for pth in save_paths:
        if not os.path.exists(pth):
            os.makedirs(pth)

    ### Seed setting for random operations
    seed = 1234
    torch.manual_seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    if cuda_check:
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    # data loaders
    print("\n===> Loading Dataset")

    _, _, testloader = get_illuminant_est_datset(args)

    print("\n==> Model initialization")

    model = SimSiam_illest(args)

    criterion = RecoveryLoss()

    ### GPU setting

    print("\n===> Setting GPU")

    if cuda_check:
        model.to(args.device)
        criterion.to(args.device)

    ############################################################################
    ############################################################################
    ############################################################################
    ##### TEST

    log_file = open(save_path + "/test.csv", "w")
    wr = csv.writer(log_file, quoting=csv.QUOTE_ALL)

    state = torch.load(args.ckpt)
    model.load_state_dict(state["model_state_dict"])

    errors = []
    ii = 1
    with torch.no_grad():
        for data in testloader:
            print("Testing image {}/{}".format(ii, len(testloader)), end="\r")
            ii += 1
            images, targets, name = data
            if cuda_check:
                images = images.to(args.device)
                targets = targets.to(args.device)

            outputs = model(images)

            error = 180 * criterion(outputs, targets).item() / math.pi
            errors.append(error)

            results = [
                name[0],
                targets[0, 0].item(),
                targets[0, 1].item(),
                targets[0, 2].item(),
                outputs[0, 0].item(),
                outputs[0, 1].item(),
                outputs[0, 2].item(),
                error,
            ]

            wr.writerow(results)
            log_file.flush()

    print(
        "\nTest finished:\n\tangular error\n\tAGV: {}\n\tMedian: {}\n\tMax: {}".format(
            st.mean(errors), st.median(errors), max(errors)
        )
    )
