import csv
import math
import os
import random
import statistics as st

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

from data.dataloaders import get_illuminant_est_datset
from models.simsiam import SimSiam_illest
from utils.ColorMetrics import RecoveryLoss
from utils.option_parser import optionParser
from utils.diagnostic import plot_grad_flow

if __name__ == "__main__":

    ### Training Options

    torch.backends.cudnn.deterministic = True

    OP = optionParser()

    OP.parser = SimSiam_illest.add_model_specific_args(OP.parser)

    args = OP.get_args()

    ### cuda setup
    cuda_check = False
    if torch.cuda.is_available() and args.device != "cpu":
        cuda_check = True

    ## Checkpointing + Save dir setup

    dirs = []
    save_dir = args.save_dir + "/" + args.arch + "_" + args.dataset + "/"
    dirs.append(save_dir)

    for d in dirs:
        if not os.path.exists(d):
            os.makedirs(d)

    if args.from_scratch:
        save_path = save_dir + "version_FS_"
    else:
        save_path = save_dir + "version_FT_" + args.base_db + "_"

    iv = 0
    while os.path.exists(save_path + str(iv)):
        iv += 1

    save_path = save_path + str(iv)

    check_dir = save_path + "/checkpoints/"

    save_paths = [save_path, check_dir]

    for pth in save_paths:
        if not os.path.exists(pth):
            os.makedirs(pth)

    log_file = open(save_path + "/log.csv", "w")
    wr = csv.writer(log_file, quoting=csv.QUOTE_ALL)

    print(
        "\n===> Options\n\nNumber of epochs: {}\nBatch size: {}\nLearning Rate: {}\nGPU device: {}\nSave directory: {}".format(
            args.epochs, args.batch_size, args.lr, args.device, save_path
        )
    )

    ### Seed setting for random operations
    seed = 1234
    torch.manual_seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    if cuda_check:
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    # data loaders
    print("\n===> Loading Dataset")

    trainloader, valloader, testloader = get_illuminant_est_datset(args)

    print("\n==> Model initialization")

    model = SimSiam_illest(args)

    # checkpoint loading
    if not args.from_scratch and args.ckpt != None:
        print("\n==> Resuming from checkpoint")
        checkpoint = torch.load(args.ckpt)

        if "model_state_dict" in checkpoint:
            model.load_pretrained(checkpoint["model_state_dict"])
        elif "state_dict" in checkpoint:
            model.load_pretrained(checkpoint["state_dict"])
        else:
            model.load_pretrained(checkpoint)

    if not args.from_scratch and not args.train_all:
        model.freeze_encoder()
        print("\nEncoder on eval.")
        model.encoder.eval()

    best_angle = 10000000

    # optimizer + loss
    criterion = RecoveryLoss()

    if args.from_scratch or args.train_all:
        if args.optim == "sgd":
            print("\nOptimizer on all:: SGD")
            optimizer = optim.SGD(
                model.parameters(),
                lr=args.lr,
                momentum=0.9,
                weight_decay=args.weight_decay,
            )
        elif args.optim == "adam":
            print("\nOptimizer on all:: Adam")
            optimizer = optim.Adam(
                model.parameters(), lr=args.lr, weight_decay=args.weight_decay
            )
        else:
            raise Exception("Options: optimizer not defined")
    else:
        if args.optim == "sgd":
            print("\nOptimizer on evaluator:: SGD")
            optimizer = optim.SGD(
                model.evaluator.parameters(),
                lr=args.lr,
                momentum=0.9,
                weight_decay=args.weight_decay,
            )
        elif args.optim == "adam":
            print("\nOptimizer on evaluator:: Adam")
            optimizer = optim.Adam(
                model.evaluator.parameters(), lr=args.lr, weight_decay=args.weight_decay
            )
        else:
            raise Exception("Options: optimizer not defined")

    if args.scheduler:
        T_max = len(trainloader) * args.epochs
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=T_max)

    ### GPU setting

    print("\n===> Setting GPU")

    if cuda_check:
        model.to(args.device)
        criterion.to(args.device)

    ## Training Loop

    print("\n===> Training\n")
    for epoch in range(args.epochs):
        info_str = []

        running_loss = 0.0
        mean_tr_loss = 0.0

        for i, batch in enumerate(trainloader):

            # zero the parameter gradients
            optimizer.zero_grad()

            inputs = batch[0]
            targets = batch[1]

            if cuda_check:
                inputs = inputs.to(args.device)
                targets = targets.to(args.device)

            outputs = model(inputs)

            loss = criterion(outputs, targets)

            loss.backward()
            optimizer.step()

            # print statistics
            running_loss += loss.item()
            mean_tr_loss += loss.item()

            if (
                i % 10 == 9 or i == len(trainloader) - 1
            ):  # print every 2000 mini-batches

                div = 10
                if i == 0:
                    div = 1
                elif i == len(trainloader) - 1:
                    div = len(trainloader) % 10
                print(
                    "[{}, {}/{}] loss: {:.8f}".format(
                        epoch + 1, i + 1, len(trainloader), running_loss / div
                    ),
                    end="\r",
                )
                # plot_grad_flow(model.named_parameters())
                running_loss = 0.0

        info_str = [epoch, mean_tr_loss / len(trainloader)]

        mean_angle = 0
        with torch.no_grad():
            for data in valloader:
                images, targets = data
                if cuda_check:
                    images = images.to(args.device)
                    targets = targets.to(args.device)

                outputs = model(images)

                mean_angle += criterion(outputs, targets)

        mean_angle = mean_angle.item() / len(valloader)

        ## Save + Logging
        if mean_angle < best_angle:

            test = os.listdir(check_dir)

            for item in test:
                if item.endswith(".pt"):
                    os.remove(os.path.join(check_dir, item))

            state = {
                "epoch": epoch,
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
                "loss": mean_tr_loss / len(trainloader),
                "angular error": mean_angle,
                "hyperparams": args,
            }
            torch.save(
                state, check_dir + "/epoch" + str(epoch) + "_resnet18_102flowers.pt"
            )
            best_angle = mean_angle

        print(
            "Mean angular error of the network on epoch {} test images: {:.8f}".format(
                epoch, 180 * mean_angle / math.pi
            )
        )
        info_str += [mean_angle]

        wr.writerow(info_str)
        log_file.flush()

        if args.scheduler:
            scheduler.step()

    log_file.close()
    print("Finished Training")

    ############################################################################
    ############################################################################
    ############################################################################
    ##### TEST

    print(
        "\nBest model:\n\tepoch:{}\n\tval angle:{}\n".format(
            state["epoch"], state["angular error"]
        )
    )

    log_file = open(save_path + "/test.csv", "w")
    wr = csv.writer(log_file, quoting=csv.QUOTE_ALL)

    model.load_state_dict(state["model_state_dict"])

    errors = []
    ii = 1
    with torch.no_grad():
        for data in testloader:
            print("Testing image {}/{}".format(ii, len(testloader)), end="\r")
            ii += 1
            images, targets, name = data
            if cuda_check:
                images = images.to(args.device)
                targets = targets.to(args.device)

            outputs = model(images)

            error = 180 * criterion(outputs, targets).item() / math.pi
            errors.append(error)

            results = [
                name[0],
                targets[0, 0].item(),
                targets[0, 1].item(),
                targets[0, 2].item(),
                outputs[0, 0].item(),
                outputs[0, 1].item(),
                outputs[0, 2].item(),
                error,
            ]

            wr.writerow(results)
            log_file.flush()

    print(
        "\nTest finished:\n\tangular error\n\tAGV: {}\n\tMedian: {}\n\tMax: {}".format(
            st.mean(errors), st.median(errors), max(errors)
        )
    )
