import csv
import os
import random
import time

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader

from data.dataloaders import get_training_data
from models.simsiam import FT_SimSiam, FT_BarlowTwins
from utils.option_parser import optionParser
import warnings

warnings.filterwarnings("ignore", "(Possibly )?corrupt EXIF data", UserWarning)

if __name__ == "__main__":

    # Training Options

    torch.backends.cudnn.deterministic = True

    OP = optionParser()

    OP.parser = FT_SimSiam.add_model_specific_args(OP.parser)

    args = OP.get_args()

    # cuda setup
    cuda_check = False
    if torch.cuda.is_available() and args.device != "cpu":
        cuda_check = True

    print("=> Cuda check: {}".format(cuda_check))

    # Checkpointing + Save dir setup

    dirs = []
    save_dir = args.save_dir + "/" + args.arch + "_" + args.dataset + "/"
    dirs.append(save_dir)

    for d in dirs:
        if not os.path.exists(d):
            os.makedirs(d)

    if args.from_scratch:
        save_path = save_dir + "version_FS_"
    else:
        save_path = save_dir + "version_FT_" + args.base_db + "_"

    iv = 0
    while os.path.exists(save_path + str(iv)):
        iv += 1

    save_path = save_path + str(iv)

    check_dir = save_path + "/checkpoints/"

    save_paths = [save_path, check_dir]

    for pth in save_paths:
        if not os.path.exists(pth):
            os.makedirs(pth)

    log_file = open(save_path + "/log.csv", "w")
    wr = csv.writer(log_file, quoting=csv.QUOTE_ALL)

    print(
        "\n===> Options\n\nNumber of epochs: {}\nBatch size: {}\nLearning Rate: {}\nGPU"
        " device: {}\nSave directory: {}".format(
            args.epochs, args.batch_size, args.lr, args.device, save_path
        )
    )

    # Seed setting for random operations
    seed = 1234
    torch.manual_seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    if cuda_check:
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    # data loaders
    print("\n===> Loading Dataset")

    trainloader, valloader, n_classes = get_training_data(args)

    # add model specific args.
    print("\n==> Model initialization")

    # if args.model == "barlowtwins":
    #     model = FT_BarlowTwins(args, n_classes)
    # else:
    model = FT_SimSiam(args, n_classes)

    # print(model)

    # checkpoint loading
    if not args.from_scratch and args.ckpt != None:
        print("\n==> Resuming from checkpoint")
        checkpoint = torch.load(
            args.ckpt, map_location=lambda storage, loc: storage)

        if "model_state_dict" in checkpoint:
            model.load_pretrained(checkpoint["model_state_dict"])
        elif "state_dict" in checkpoint:
            model.load_pretrained(checkpoint["state_dict"])
        else:
            model.load_pretrained(checkpoint)
        if not args.from_scratch and not args.train_all:
            print("\nEncoder on eval.")
            model.encoder.eval()

    best_accuracy = 0

    # optimizer + loss
    criterion = nn.CrossEntropyLoss()

    if args.from_scratch or args.train_all:
        if args.optim == "sgd":
            print("\nOptimizer on all:: SGD")
            optimizer = optim.SGD(
                model.parameters(),
                lr=args.lr,
                momentum=0.9,
                weight_decay=args.weight_decay,
            )
        elif args.optim == "adam":
            print("\nOptimizer on all:: Adam")
            optimizer = optim.Adam(
                model.parameters(), lr=args.lr, weight_decay=args.weight_decay
            )
        else:
            raise Exception("Options: optimizer not defined")
    else:
        if args.optim == "sgd":
            print("\nOptimizer on evaluator:: SGD")
            optimizer = optim.SGD(
                model.evaluator.parameters(),
                lr=args.lr,
                momentum=0.9,
                weight_decay=args.weight_decay,
            )
        elif args.optim == "adam":
            print("\nOptimizer on evaluator:: Adam")
            optimizer = optim.Adam(
                model.evaluator.parameters(), lr=args.lr, weight_decay=args.weight_decay
            )
        else:
            raise Exception("Options: optimizer not defined")

    if args.scheduler:
        T_max = len(trainloader) * args.epochs
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=T_max)

    # GPU setting

    print("\n===> Setting GPU")

    if cuda_check:
        model.to(args.device)
        criterion.to(args.device)

    # Training Loop

    print("\n===> Training\n")
    for epoch in range(args.epochs):
        info_str = []

        running_loss = 0.0
        mean_tr_loss = 0.0

        model.train()
        correct = 0
        total = 0
        if not args.from_scratch and not args.train_all:
            model.encoder.eval()
        for i, batch in enumerate(trainloader):

            st = time.time()
            # zero the parameter gradients
            optimizer.zero_grad()

            inputs = batch[0]
            labels = batch[1]

            if cuda_check:
                inputs = inputs.to(args.device)
                labels = labels.to(args.device)

            outputs = model(inputs)

            loss = criterion(outputs, labels)

            _, predicted = torch.max(outputs.data, 1)
            total += labels.size(0)
            correct += (predicted == labels).sum().item()

            loss.backward()
            optimizer.step()

            end = time.time() - st

            # print statistics
            running_loss += loss.item()
            mean_tr_loss += loss.item()

            # if (
            #     i % 10 == 9 or i == len(trainloader) - 1
            # ):  # print every 2000 mini-batches

            div = 10
            if i == 0:
                div = 1
            elif i == len(trainloader) - 1:
                div = len(trainloader) % 10
            print(
                "[{}, {}/{}] loss: {:.8f}\ttime: {:.5f}".format(
                    epoch + 1, i +
                    1, len(trainloader), running_loss / div, end
                ),
                end="\r",
            )
            running_loss = 0.0

        accuracy_train = 100 * correct / total

        info_str = [epoch, mean_tr_loss / len(trainloader)]

        correct = 0
        total = 0
        with torch.no_grad():
            model.eval()
            for data in valloader:
                images, labels = data

                if cuda_check:
                    images = images.to(args.device)
                    labels = labels.to(args.device)

                outputs = model(images)

                _, predicted = torch.max(outputs.data, 1)
                total += labels.size(0)
                correct += (predicted == labels).sum().item()

        accuracy = 100 * correct / total

        ## Save + Logging
        if accuracy > best_accuracy:

            test = os.listdir(check_dir)

            for item in test:
                if item.endswith(".pt"):
                    os.remove(os.path.join(check_dir, item))

            state = {
                "epoch": epoch,
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
                "loss": mean_tr_loss / len(trainloader),
                "accuracy": accuracy,
                "hyperparams": args,
            }
            torch.save(
                state,
                check_dir + "/epoch" +
                str(epoch) + "_resnet18_" + args.dataset + ".pt",
            )
            best_accuracy = accuracy

        print(
            "Accuracy of the network on epoch {} train images: {:.2f}%\ttest images:"
            " {:.2f}%".format(epoch, accuracy_train, accuracy)
        )
        info_str += [accuracy]

        wr.writerow(info_str)
        log_file.flush()

        if args.scheduler:
            scheduler.step()

    log_file.close()
    print("Finished Training")
